"""
adaptive split data into bins.
"""

import functools

import numpy as np

from .data import data_index, data_mask

try:
    import matplotlib.patches as mpathes
except ImportError:
    pass


class AdaptiveBound(object):
    """adaptive bound cut for data value"""

    def __init__(self, base_data, bins):
        if isinstance(bins, int):
            self._base_data = np.array([base_data])
            self.bins = [bins]
        elif isinstance(bins, list):
            self._base_data = np.array(base_data)
            self.bins = bins
        else:
            raise TypeError(
                "bins should be int or list of int, "
                + "insteading of {}".format(type(bins))
            )

    @functools.lru_cache()
    def get_bounds_data(self):
        """get split data bounds, and the data after splitting"""
        base_bound = AdaptiveBound.base_bound(self._base_data)
        bounds, datas = AdaptiveBound.loop_split_bound(
            self._base_data, self.bins
        )
        return bounds, datas

    def get_bounds(self):
        """get split data bounds"""
        bounds, _ = self.get_bounds_data()
        return bounds

    def get_bool_mask(self, data):
        """bool mask for splitting data"""
        bounds = self.get_bounds()
        idx_data = data[: bounds[0][0].shape[0]]
        ret = []
        for lb, rb in bounds:
            mask = np.logical_and(
                idx_data >= lb[..., np.newaxis], idx_data < rb[..., np.newaxis]
            )
            mask = np.all(mask, axis=0)
            ret.append(mask)
        return ret

    def split_full_data(self, data, base_index=None):
        """split structure data, (TODO because large IO,  the method is slow.)"""
        base_data = [[data_index(data, i) for i in base_index]]
        mask = self.get_bool_mask(base_data)
        ret = []
        for i in mask:
            ret.append(data_mask(data, i))
        return ret

    def split_data(self, data):
        """split data, the shape is same as base_data"""
        mask = self.get_bool_mask(data)
        ret = []
        for i in mask:
            ret.append(data[..., i])
        return ret

    @staticmethod
    def single_split_bound(data, n=2, base_bound=None):
        """split data in the order of data value

        >>> data = np.array([1.0, 2.0, 1.4, 3.1])
        >>> AdaptiveBound.single_split_bound(data)
        [(1.0, 1.7...), (1.7..., 3.1...)]

        """
        if base_bound is None:
            base_bound = np.min(data), np.max(data) + 1e-6
        num_lb = base_bound[0]
        bounds = []
        for j in range(1, n):
            num_rb = np.percentile(data, j / n * 100, axis=0) + 1e-6
            bounds.append((num_lb, num_rb))
            num_lb = num_rb
        bounds.append((num_lb, base_bound[1]))
        return bounds

    @staticmethod
    def multi_split_bound(datas, n, base_bound=None):
        """multi data for single_split_bound, so `n` is list of int

        >>> data = np.array([[1.0, 2.0, 1.4, 3.1], [2.0, 1.0, 3.0, 1.0]])
        >>> bound, _ = AdaptiveBound.multi_split_bound(data, [2, 1])
        >>> [(i[0][0]+1e-6, i[1][0]+1e-6) for i in bound]
        [(1.0..., 1.7...), (1.7..., 3.1...)]

        """
        datas = np.array(datas)
        if base_bound is None:
            base_bound = AdaptiveBound.base_bound(datas)
        bound_chain = [base_bound]
        data_chain = [datas]
        for idx, size in enumerate(n):
            new_bound_chain = []
            new_data_chain = []
            for bnd, data in zip(bound_chain, data_chain):
                idx_data = data[idx]
                bounds = AdaptiveBound.single_split_bound(
                    idx_data, size, base_bound=(bnd[0][idx], bnd[1][idx])
                )
                for i in bounds:
                    lb, rb = i
                    l_bnd, r_bnd = bnd
                    l_bnd = np.copy(l_bnd)
                    r_bnd = np.copy(r_bnd)
                    l_bnd[idx] = lb
                    r_bnd[idx] = rb
                    new_bound_chain.append((l_bnd, r_bnd))
                    mask = np.logical_and(idx_data >= lb, idx_data < rb)
                    new_data_chain.append(data[:, mask])
            bound_chain = new_bound_chain
            data_chain = new_data_chain
        return bound_chain, data_chain

    @staticmethod
    def loop_split_bound(datas, n, base_bound=None):
        """loop for multi_split_bound, so `n` is list of list of int"""
        datas = np.array(datas)
        if base_bound is None:
            base_bound = AdaptiveBound.base_bound(datas)
        bound_chain = [base_bound]
        data_chain = [datas]
        for idx, size in enumerate(n):
            new_bound_chain = []
            new_data_chain = []
            for bnd, data in zip(bound_chain, data_chain):
                bound, data_i = AdaptiveBound.multi_split_bound(
                    data, size, base_bound=bnd
                )
                new_bound_chain += bound
                new_data_chain += data_i
            bound_chain = new_bound_chain
            data_chain = new_data_chain
        return bound_chain, data_chain

    @staticmethod
    def base_bound(data):
        """base bound for the data"""
        lb = np.min(data, axis=-1) - 1e-6
        rb = np.max(data, axis=-1) + 1e-6
        return (lb, rb)

    def get_bound_patch(self, **kwargs):
        ret = []
        for i, bnd in enumerate(self.get_bounds()):
            min_x, min_y = bnd[0]
            max_x, max_y = bnd[1]
            rect = mpathes.Rectangle(
                (min_x, min_y), max_x - min_x, max_y - min_y, **kwargs
            )  # cmap(weights[i]/max_weight))
            ret.append(rect)
        return ret

    def plot_bound(self, ax, **kwargs):
        for i in self.get_bound_patch(**kwargs):
            ax.add_patch(i)


def cal_chi2(numbers, n_fp):
    weights = []
    # print(numbers)
    # chi21 = []
    for ndata, nmc in numbers:
        weight = (ndata - nmc) / np.sqrt(np.abs(ndata))
        weights.append(weight ** 2)
        # chi21.append(ndata * np.log(nmc))
    max_weight = np.max(weights)
    chi2 = np.sum(weights)
    print("bins: ", len(weights))
    print("number of free parameters: ", n_fp)
    ndf = len(weights) - 1 - n_fp
    print(
        "chi2/ndf: ", np.sum(weights), "/", ndf
    )  # ,"another", np.sum(chi21))
    return chi2, ndf



#!/usr/bin/env python3

import datetime

# from pprint import pprint
import json
import os.path
import sys
import time

import numpy as np
from scipy.optimize import minimize

import tf_pwa
from tf_pwa.amp import AmplitudeModel, DecayGroup, HelicityDecay, Particle
from tf_pwa.cal_angle import prepare_data_from_decay
from tf_pwa.data import data_to_tensor, split_generator
from tf_pwa.fitfractions import cal_fitfractions
from tf_pwa.model import FCN, CombineFCN, Model
from tf_pwa.tensorflow_wrapper import tf
from tf_pwa.utils import error_print, load_config_file, pprint
from tf_pwa.variable import VarsManager

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


# log_dir = "./cached_dir/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")


data_file_name = [
    ["./data/data4600_new.dat", "data/Dst0_data4600_new.dat"],
    ["./data/bg4600_new.dat", "data/Dst0_bg4600_new.dat"],
    ["./data/PHSP4600_new.dat", "data/Dst0_PHSP4600_new.dat"],
]

data_file_name2 = [
    ["./data/data4600_new.dat", "data/Dst0_data4600_new.dat"],
    ["./data/bg4600_new.dat", "data/Dst0_bg4600_new.dat"],
    ["./data/PHSP4600_new.dat", "data/Dst0_PHSP4600_new.dat"],
]


def prepare_data(fnames, decs, particles=None, dtype="float64"):
    tname = ["data", "bg", "PHSP"]
    data_np = {}
    for i, name in enumerate(fnames):
        data_np[tname[i]] = prepare_data_from_decay(
            name[0], decs, particles=particles, dtype=dtype
        )
    data, bg, mcdata = data_to_tensor([data_np[i] for i in tname])
    return data, bg, mcdata


def cal_hesse_error(amp, val, w_bkg, data, mcdata, bg, args_name, batch):
    a_h = FCN(Model(amp, w_bkg), data, mcdata, bg=bg, batch=batch)
    t = time.time()
    nll, g, h = a_h.nll_grad_hessian(
        val
    )  # data_w,mcdata,weight=weights,batch=50000)
    print("Time for calculating errors:", time.time() - t)
    inv_he = np.linalg.pinv(h.numpy())
    np.save("error_matrix.npy", inv_he)
    return inv_he


def get_decay_chains(config_list):
    a = Particle("A", J=1, P=-1, spins=(-1, 1))
    b = Particle("B", J=1, P=-1)
    c = Particle("C", J=0, P=-1)
    d = Particle("D", J=1, P=-1)
    decay = {}
    for i in config_list:
        config = config_list[i]
        res = Particle(
            i,
            config["J"],
            config["Par"],
            mass=config["m0"],
            width=config["g0"],
        )
        chain = config["Chain"]
        if chain < 0:
            dec1 = HelicityDecay(a, [res, c])
            dec2 = HelicityDecay(res, [b, d])
        elif chain < 100:
            dec1 = HelicityDecay(a, [res, d])
            dec2 = HelicityDecay(res, [b, c])
        elif chain < 200:
            dec1 = HelicityDecay(a, [res, b])
            dec2 = HelicityDecay(res, [d, c])
        else:
            raise Exception("unknown chain")
        decay[i] = [dec1, dec2]

    decs = DecayGroup(a.chain_decay())
    return decs, [d, b, c], decay


def get_amplitude(decs, config_list, decay, polar=True, vm=None):
    amp = AmplitudeModel(decs, polar, vm=vm)
    for i in config_list:
        if "coef_head" in config_list[i]:
            coef_head = config_list[i]["coef_head"]
            for a, b in zip(decay[i], decay[coef_head]):
                b.g_ls.sameas(a.g_ls)
            for j in decs.chains:
                if decay[i][0] in j:
                    for k in decs.chains:
                        if decay[coef_head][0] in k:
                            k.total.r_shareto(j.total)
        if "total" in config_list[i]:
            for j in decs.chains:
                if decay[i][0] in j:
                    j.total.fixed(config_list[i]["total"])
        if ("float" in config_list[i]) and config_list[i]["float"]:
            if "m" in config_list[i]["float"]:
                decay[i][1].core.mass.freed()
            if "g" in config_list[i]["float"]:
                decay[i][1].core.width.freed()
    return amp


def load_params(amp, init_params):
    try:
        with open(init_params) as f:
            param_0 = json.load(f)
            if "value" in param_0:
                param_0 = param_0["value"]
            param = {}
            for k, v in param_0.items():
                if k.endswith(":0"):
                    k = k[:-2]
                param[k] = v
            print("using {}".format(init_params))
            amp.set_params(param)
        RDM_INI = False
    except Exception as e:
        print(e)
        RDM_INI = True
        print("using RANDOM parameters")


def fit(method="BFGS", init_params="init_params.json", hesse=True, frac=True):
    POLAR = True  # fit in polar coordinates. should be consistent with init_params.json if any
    GEN_TOY = False  # use toy data (mcdata and bg stay the same). REMEMBER to update gen_params.json

    dtype = "float64"
    w_bkg = 0.768331
    # set_gpu_mem_growth()
    tf.keras.backend.set_floatx(dtype)
    # open Resonances list as dict
    config_list = load_config_file("Resonances")
    config_list2 = load_config_file("Resonances")

    vm = VarsManager()

    decs, final_particles, decay = get_decay_chains(config_list)
    data, bg, mcdata = prepare_data(
        data_file_name, decs, particles=final_particles, dtype=dtype
    )
    decs2, final_particles2, decay2 = get_decay_chains(config_list)
    data2, bg2, mcdata2 = prepare_data(
        data_file_name2, decs2, particles=final_particles2, dtype=dtype
    )

    amp = get_amplitude(decs, config_list, decay, polar=POLAR, vm=vm)
    amp2 = get_amplitude(decs2, config_list2, decay2, polar=POLAR, vm=vm)

    load_params(amp, init_params)
    load_params(amp2, init_params)

    model = Model(amp, w_bkg=w_bkg)
    model2 = Model(amp2, w_bkg=w_bkg)
    pprint(model.get_params())

    fcn = CombineFCN(
        [model, model2],
        [data, data2],
        [mcdata, mcdata2],
        bg=[bg, bg2],
        batch=65000,
    )

    # fit configure
    args = {}
    args_name = vm.trainable_vars
    x0 = []
    bnds = []
    bounds_dict = {
        # "Zc_4160_m:0":(4.1,4.22),
        # "Zc_4160_g:0":(0,None)
    }

    for i in model.Amp.trainable_variables:
        args[i.name] = i.numpy()
        x0.append(i.numpy())
        if i.name in bounds_dict:
            bnds.append(bounds_dict[i.name])
        else:
            bnds.append((None, None))
        args["error_" + i.name] = 0.1

    points = []
    nlls = []
    now = time.time()
    maxiter = 1000

    method = "BFGS"

    def callback(x):
        if np.fabs(x).sum() > 1e7:
            x_p = dict(zip(args_name, x))
            raise Exception("x too large: {}".format(x_p))
        points.append(vm.get_all_val())
        nlls.append(float(fcn.cached_nll))
        print(fcn.cached_nll)

    vm.set_bound(bounds_dict)
    f_g = vm.trans_fcn_grad(fcn.nll_grad)
    s = minimize(
        f_g,
        np.array(vm.get_all_val(True)),
        method=method,
        jac=True,
        callback=callback,
        options={"disp": 1, "gtol": 1e-4, "maxiter": maxiter},
    )
    xn = vm.get_all_val()
    print("########## fit state:")
    print(s)
    print("\nTime for fitting:", time.time() - now)
    model.Amp.vm.trans_params(POLAR)
    val = {k: v.numpy() for k, v in model.Amp.variables.items()}
    with open("final_params.json", "w") as f:
        json.dump({"value": val}, f, indent=2)

    err = {}
    inv_he = cal_hesse_error(
        model.Amp, val, w_bkg, data, mcdata, bg, args_name, batch=20000
    )
    diag_he = inv_he.diagonal()
    hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
    print(hesse_error)
    err = dict(zip(model.Amp.vm.trainable_vars, hesse_error))
    print("\n########## fit results:")
    for i in val:
        print("  ", i, ":", error_print(val[i], err.get(i, None)))

    outdic = {"value": val, "error": err, "config": config_list}
    with open("final_params.json", "w") as f:
        json.dump(outdic, f, indent=2)

    err_frac = {}

    frac, grad = cal_fitfractions(
        model.Amp, list(split_generator(mcdata, 25000))
    )

    for i in frac:
        err_frac[i] = np.sqrt(np.dot(np.dot(inv_he, grad[i]), grad[i]))

    print("########## fit fractions")
    for i in frac:
        print(i, ":", error_print(frac[i], err_frac.get(i, None)))
    print("\nEND\n")
    """outdic = {"value": val, "error": err, "frac": frac, "err_frac": err_frac}
    with open("glbmin_params.json", "w") as f:
        json.dump(outdic, f, indent=2)
    """


def main():
    import argparse

    parser = argparse.ArgumentParser(description="simple fit scripts")
    parser.add_argument(
        "--no-GPU", action="store_false", default=True, dest="has_gpu"
    )
    parser.add_argument("--method", default="BFGS", dest="method")
    results = parser.parse_args()
    if results.has_gpu:
        with tf.device("/device:GPU:0"):
            fit(method=results.method)
    else:
        with tf.device("/device:CPU:0"):
            fit(method=results.method)


if __name__ == "__main__":

    # summary_writer = tf.summary.create_file_writer(log_dir)
    # with summary_writer.as_default():
    main()



from .config_loader import (
    ConfigLoader,
    FitResult,
    PlotParams,
    validate_file_name,
)
from .extra import *
from .multi_config import MultiConfig
from .plot import export_legend, hist_error, hist_line
from .sample import single_sampling



"""

Module for factor system.

```
A = a1 ( B x C x D) + a2 (E x F)
B = b1 B1 + b2 B2
```

is a tree structure
```
A -> [(a1, [(b1, B1), (b2, B2)], C, D), (a2, E, F)]
```

Each component is a path for root to a leaf.
```
(a1, b1),  (a1, b2),  (a2,)
```

We can add some options to change the possible combination. (TODO)

"""
import contextlib
import itertools

from tf_pwa.variable import _shape_func


def get_all_chain(a):
    for i in a:
        for j in get_prod_chain(i):
            yield j


def get_split_chain(a):
    for i in a:
        for j in get_prod_chain(i):
            yield from j


def get_prod_chain(i):
    ret = []
    for j in i:
        if isinstance(j, (list, tuple)):
            ret.append(list(get_split_chain(j)))
        else:
            ret.append([j])
    for i in itertools.product(*ret):
        yield i


def get_chain_name(chain):
    ret = []
    for i in chain:
        tmp = []
        if i.shape:
            if i.cplx:

                def fun(name, idx):
                    tmp.append((name + "r", name + "i"))

            else:

                def fun(name, idx):
                    tmp.append(name)

            _shape_func(fun, i.shape, i.name)
        else:
            if i.cplx:
                tmp.append((i.name + "r", i.name + "i"))
            else:
                tmp.append(i.name)
        ret.append(tmp)
    return itertools.product(*ret)


@contextlib.contextmanager
def temp_var(vm):
    params = vm.get_all_dic()
    yield vm
    vm.set_all(params)


def flatten_all(x):
    ret = []
    if isinstance(x, (tuple, list)):
        for i in x:
            ret += list(flatten_all(i))
        return ret
    return [x]


def get_all_partial_amp(amp, data, strip_part=[]):
    var = amp.decay_group.get_factor_variable()
    chains = list(get_all_chain(var))

    part = []
    for i in chains:
        part += list(get_chain_name(i))

    ret_part = []
    for j in strip_variable(part, strip_part):
        ret_part.append(j)
    all_var = flatten_all(ret_part)

    ret = []
    id_ = []
    for i in ret_part:
        ret.append(partial_amp(amp, data, all_var, flatten_all(i)))
        id_.append(get_id_variable(all_var, i))
    return id_, ret


def partial_amp(amp, data, all_va, need_va):
    with temp_var(amp.vm) as vm:
        others_va = set(all_va) - set(need_va)
        amp.vm.set_all({i: 0.0 for i in others_va})
        ret = amp(data)
    return ret


def strip_variable(var_all, part=[]):
    def match(x):
        for i in part:
            if i in x:
                return True
        return False

    cached_list = [frozenset()]

    for j in var_all:
        j = flatten_all(j)
        after_strip = [k for k in j if not match(k)]
        if frozenset(after_strip) in cached_list:
            continue
        yield after_strip
        cached_list.append(frozenset(after_strip))


def get_id_variable(all_var, var):
    var = flatten_all(var)
    ret = frozenset(set(var) - (set(all_var) - set(var)))
    if len(ret) == 0:
        return frozenset(var)
    return ret



import numpy as np

from tf_pwa.fit_improve import *


def rosen(x):
    """The Rosenbrock function"""
    return sum(100.0 * (x[1:] - x[:-1] ** 2.0) ** 2.0 + (1 - x[:-1]) ** 2.0)


def rosen_der(x):
    xm = x[1:-1]
    xm_m1 = x[:-2]
    xm_p1 = x[2:]
    der = np.zeros_like(x)
    der[1:-1] = (
        200 * (xm - xm_m1 ** 2) - 400 * (xm_p1 - xm ** 2) * xm - 2 * (1 - xm)
    )
    der[0] = -400 * x[0] * (x[1] - x[0] ** 2) - 2 * (1 - x[0])
    der[-1] = 200 * (x[-1] - x[-2] ** 2)
    # der += np.random.random(der.shape)
    return der


def rosen_hess(x):
    x = np.asarray(x)
    H = np.diag(-400 * x[:-1], 1) - np.diag(400 * x[:-1], -1)
    diagonal = np.zeros_like(x)
    diagonal[0] = 1200 * x[0] ** 2 - 400 * x[1] + 2
    diagonal[-1] = 200
    diagonal[1:-1] = 202 + 1200 * x[1:-1] ** 2 - 400 * x[2:]
    H = H + np.diag(diagonal)
    return H


def test_minimize():
    def my_fit_fun(x):
        return rosen(x), rosen_der(x)

    x0 = np.array([2.0, 1.3, 0.7, 0.8, 1.9, 1.2])
    ret = minimize(my_fit_fun, x0)
    print(ret)
    assert np.allclose(ret.x, np.ones_like(x0))



#!/usr/bin/env python3

import json
import math
import os
import os.path
import subprocess
import sys
import threading
import time
from sys import platform

import numpy as np
import pymultinest
import tensorflow as tf
from numpy import cos, pi
from pymultinest.solve import solve
from scipy.optimize import BFGS, basinhopping, minimize

from tf_pwa.amplitude import AllAmplitude, param_list
from tf_pwa.angle import cal_ang_file, cal_ang_file4
from tf_pwa.bounds import Bounds
from tf_pwa.fitfractions import cal_fitfractions
from tf_pwa.model import FCN, Cache_Model, param_list
from tf_pwa.utils import error_print, flatten_np_data, load_config_file, pprint

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


try:
    os.mkdir("chains")
except OSError:
    pass


def show(filepath):
    """ open the output (pdf) file for the user """
    if os.name == "mac" or platform == "darwin":
        subprocess.call(("open", filepath))
    elif os.name == "nt" or platform == "win32":
        os.startfile(filepath)
    elif platform.startswith("linux"):
        subprocess.call(("xdg-open", filepath))


def prepare_data(dtype="float64", model="3"):
    fname = [
        ["./data/data4600_new.dat", "data/Dst0_data4600_new.dat"],
        ["./data/bg4600_new.dat", "data/Dst0_bg4600_new.dat"],
        ["./data/PHSP4600_new.dat", "data/Dst0_PHSP4600_new.dat"],
    ]
    tname = ["data", "bg", "PHSP"]
    data_np = {}
    for i in range(len(tname)):
        if model == "3":
            data_np[tname[i]] = cal_ang_file(fname[i][0], dtype)
        elif model == "4":
            data_np[tname[i]] = cal_ang_file4(fname[i][0], fname[i][1], dtype)

    def load_data(name):
        dat = []
        tmp = flatten_np_data(data_np[name])
        for i in param_list:
            tmp_data = tf.Variable(tmp[i], name=i, dtype=dtype)
            dat.append(tmp_data)
        return dat

    # with tf.device('/device:GPU:0'):
    data = load_data("data")
    bg = load_data("bg")
    mcdata = load_data("PHSP")
    return data, bg, mcdata


def main():
    dtype = "float64"
    w_bkg = 0.768331
    # set_gpu_mem_growth()
    tf.keras.backend.set_floatx(dtype)
    # open Resonances list as dict
    config_list = load_config_file("Resonances")

    data, bg, mcdata = prepare_data(dtype=dtype)  # ,model="4")

    amp = AllAmplitude(config_list)
    a = Cache_Model(
        amp, w_bkg, data, mcdata, bg=bg, batch=65000
    )  # ,constrain={"Zc_4160_g0:0":(0.1,0.1)})
    try:
        with open("final_params.json") as f:
            param = json.load(f)
            # print("using init_params.json")
            if "value" in param:
                a.set_params(param["value"])
            else:
                a.set_params(param)
    except:
        pass
    # print(a.Amp(data))
    # exit()
    # pprint(a.get_params())
    # print(data,bg,mcdata)
    # t = time.time()
    # nll,g = a.cal_nll_gradient()#data_w,mcdata,weight=weights,batch=50000)
    # print("nll:",nll,"Time:",time.time()-t)
    # exit()
    fcn = FCN(a)

    # fit configure
    args = {}
    args_name = []
    x0 = []
    bnds = []
    bounds_dict = {"Zc_4160_m0:0": (4.1, 4.22), "Zc_4160_g0:0": (0, None)}

    for i in a.Amp.trainable_variables:
        args[i.name] = i.numpy()
        x0.append(i.numpy())
        args_name.append(i.name)
        if i.name in bounds_dict:
            bnds.append(bounds_dict[i.name])
        else:
            bnds.append((None, None))
        args["error_" + i.name] = 0.1

    bnds_cube = []
    for a, b in bnds:
        if a is None:
            if b is None:
                bnds_cube.append((200, -100))
            else:
                bnds_cube.append((b + 100, -100))
        else:
            if b is None:
                bnds_cube.append((100 - a, a))
            else:
                bnds_cube.append((b - a, a))
    # bnds_cube=[(0,1),(0,1)]
    cube_size = np.array(bnds_cube)

    def Prior(cube):
        cube[:] = cube * cube_size[:, 0] + cube_size[:, 1]
        return cube

    def LogLikelihood(cube):
        # return -np.sum(cube*cube)
        ret = fcn(cube)
        return -ret

    prefix = "chains/1-"
    with open("%sparams.json" % prefix, "w") as f:
        json.dump(args_name, f, indent=2)
    n_params = len(args_name)

    # we want to see some output while it is running
    # progress = pymultinest.ProgressPlotter(n_params = n_params, outputfiles_basename=prefix);
    # progress.start()
    # threading.Timer(30, show, [prefix+"phys_live.points.pdf"]).start() # delayed opening
    now = time.time()
    solution = solve(
        LogLikelihood,
        Prior,
        n_dims=n_params,
        outputfiles_basename=prefix,
        importance_nested_sampling=False,
        verbose=True,
    )
    print("Time for fitting:", time.time() - now)
    # progress.stop()
    # print(solution)
    print()
    print("evidence: %(logZ).1f +- %(logZerr).1f" % solution)
    print()
    print("parameter values:")
    for name, col in zip(args_name, solution["samples"].transpose()):
        print("%15s : %.3f +- %.3f" % (name, col.mean(), col.std()))
    # use: solution.logZ, solution.logZerr, solution.samples


if __name__ == "__main__":
    main()



import warnings

import numpy as np

from tf_pwa.amp import get_particle
from tf_pwa.cal_angle import cal_angle_from_momentum
from tf_pwa.config_loader.data import SimpleData, register_data_mode
from tf_pwa.data import data_shape


@register_data_mode("simple_npz")
class NpzData(SimpleData):
    def get_particle_p(self):
        order = self.dic.get("particle_p", None)
        if order is None:
            return self.get_dat_order()
        return order

    def get_data(self, idx) -> dict:
        if self.cached_data is not None:
            data = self.cached_data.get(idx, None)
            if data is not None:
                return data
        files = self.get_data_file(idx)
        weight_sign = self.get_weight_sign(idx)
        return self.load_data(files, weight_sign)

    def load_data(
        self, files, weights=None, weights_sign=1, charge=None
    ) -> dict:
        # print(files, weights)
        if files is None:
            return None
        order = self.get_dat_order()
        p_list = self.get_particle_p()
        center_mass = self.dic.get("center_mass", True)
        r_boost = self.dic.get("r_boost", False)
        random_z = self.dic.get("random_z", False)
        npz_data = np.load(files)
        p = {
            get_particle(str(v)): npz_data[str(k)]
            for k, v in zip(p_list, order)
        }
        data = cal_angle_from_momentum(
            p,
            self.decay_struct,
            center_mass=center_mass,
            r_boost=r_boost,
            random_z=random_z,
        )
        if "weight" in npz_data:
            data["weight"] = npz_data["weight"]
        if "charge_conjugation" in npz_data:
            data["charge_conjugation"] = npz_data["charge_conjugation"]
        else:
            data["charge_conjugation"] = np.ones((data_shape(data),))
        return data


@register_data_mode("multi_npz")
class MultiNpzData(NpzData):
    def __init__(self, *args, **kwargs):
        super(MultiNpzData, self).__init__(*args, **kwargs)
        self._Ngroup = 0

    def get_data(self, idx) -> list:
        if self.cached_data is not None:
            data = self.cached_data.get(idx, None)
            if data is not None:
                return data
        files = self.get_data_file(idx)
        if files is None:
            return None
        if not isinstance(files[0], list):
            files = [files]
        weight_sign = self.get_weight_sign(idx)
        ret = [self.load_data(i, weight_sign) for i in files]
        if self._Ngroup == 0:
            self._Ngroup = len(ret)
        elif idx != "phsp_noeff" and self._Ngroup != len(ret):
            warnings.warn("not the same data group")
        return ret

    def get_phsp_noeff(self):
        if "phsp_noeff" in self.dic:
            phsp_noeff = self.get_data("phsp_noeff")
            assert len(phsp_noeff) == 1
            return phsp_noeff[0]
        warnings.warn(
            "No data file as 'phsp_noeff', using the first 'phsp' file instead."
        )
        return self.get_data("phsp")[0]



"""
Examples for particle model
---------------------------

decay system is model as

DecayGroup
    DecayChain

        Decay

        Particle

"""

import matplotlib.pyplot as plt

from tf_pwa.amp import Decay, DecayChain, DecayGroup, Particle
from tf_pwa.vis import plot_decay_struct

# %%
# We can easy create some instance of Particle
# and then combine them as Decay

a = Particle("A")
b = Particle("B")
c = Particle("C")
d = Particle("D")

r = Particle("R")

dec1 = Decay(a, [r, b])
dec2 = Decay(r, [c, d])

# %%
#
# DecayChain is a list of Decays.

decay_chain = DecayChain([dec1, dec2])
decay_chain

# %%
# We can plot it using matplotlib.

plot_decay_struct(decay_chain)
plt.show()

# %%
# DecayGroup is a list of DecayChain with the same initial and final states

decay_group = DecayGroup([decay_chain])
decay_group

# %%
# We can build a simple function to infer the charge from final states.


def charge_infer(dec, charge_map):
    # use particle equal condition
    cached_charge = {Particle(k): v for k, v in charge_map.items()}
    # loop for all decays in decay chain
    for i, dec_i in dec.depth_first(False):
        # all out particles has charge
        assert all(i in cached_charge for i in dec_i.outs)
        # the charge or core particle is the sum of
        cached_charge[dec_i.core] = sum(cached_charge[i] for i in dec_i.outs)
    return cached_charge


charges = {
    "B": -1,
    "C": 0,
    "D": 1,
}

charge_infer(decay_chain, charges)

# %%
# See more in `~tf_pwa.cal_angle.cal_chain_boost`.



#!/usr/bin/env python3

import datetime

# from pprint import pprint
import json
import math
import os
import os.path
import sys
import time

import numpy as np
from scipy.optimize import BFGS, basinhopping, minimize

import tf_pwa
from tf_pwa.amp import (
    AmplitudeModel,
    DecayGroup,
    HelicityDecay,
    Particle,
    get_name,
)
from tf_pwa.cal_angle import prepare_data_from_decay
from tf_pwa.data import data_to_numpy, data_to_tensor, split_generator
from tf_pwa.fitfractions import cal_fitfractions, cal_fitfractions_no_grad
from tf_pwa.model import FCN, Model
from tf_pwa.tensorflow_wrapper import tf
from tf_pwa.utils import (
    error_print,
    flatten_np_data,
    load_config_file,
    pprint,
    std_polar,
)

this_dir = os.path.dirname(__file__)
sys.path.insert(0, this_dir + "/..")


# from tf_pwa.bounds import Bounds


log_dir = "./cached_dir/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")


def load_cached_data(cached_data_file="cached_data_new.npy"):
    cached_dir = "./cached_dir"
    if not os.path.exists(cached_dir):
        return None
    cached_path = os.path.join(cached_dir, cached_data_file)
    if os.path.exists(cached_path):
        cached_data = tf_pwa.load_data(cached_path)
        return cached_data
    return None


def save_cached_data(cached_data, cached_data_file="cached_data_new.npy"):
    cached_dir = "./cached_dir"
    if not os.path.exists(cached_dir):
        os.mkdir(cached_dir)
    cached_path = os.path.join(cached_dir, cached_data_file)
    tf_pwa.save_data(cached_path, cached_data)


def prepare_data(decs, particles=None, dtype="float64"):
    fname = [
        ["./data/data4600_new.dat", "data/Dst0_data4600_new.dat"],
        ["./data/bg4600_new.dat", "data/Dst0_bg4600_new.dat"],
        ["./data/PHSP4600_new.dat", "data/Dst0_PHSP4600_new.dat"],
    ]
    tname = ["data", "bg", "PHSP"]
    try:
        cached_data = load_cached_data()
    except Exception as e:
        print(e)
    cached_data = None
    if cached_data is not None:
        cached_data = data_to_tensor(cached_data)
        pprint(cached_data)
        data = cached_data["data"]
        bg = cached_data["bg"]
        mcdata = cached_data["PHSP"]
        print("using cached data")
        return data, bg, mcdata
    data_np = {}
    for i, name in enumerate(fname):
        data_np[tname[i]] = prepare_data_from_decay(
            name[0], decs, particles=particles, dtype=dtype
        )

    data, bg, mcdata = data_to_tensor([data_np[i] for i in tname])
    # import pprint
    # pprint.pprint(data)
    save_cached_data(data_to_numpy({"data": data, "bg": bg, "PHSP": mcdata}))
    return data, bg, mcdata


def cal_hesse_error(amp, val, w_bkg, data, mcdata, bg, args_name, batch):
    a_h = FCN(Model(amp, w_bkg), data, mcdata, bg=bg, batch=batch)
    t = time.time()
    nll, g, h = a_h.nll_grad_hessian(
        val
    )  # data_w,mcdata,weight=weights,batch=50000)
    print("Time for calculating errors:", time.time() - t)
    # print(nll)
    # print([i.numpy() for i in g])
    # print(h.numpy())
    inv_he = np.linalg.pinv(h.numpy())
    np.save("error_matrix.npy", inv_he)
    # print("edm:",np.dot(np.dot(inv_he,np.array(g)),np.array(g)))
    return inv_he


def get_decay_chains(config_list):
    a = Particle("A", J=1, P=-1, spins=(-1, 1))
    b = Particle("B", J=1, P=-1)
    c = Particle("C", J=0, P=-1)
    d = Particle("D", J=1, P=-1)
    decay = {}
    for i in config_list:
        config = config_list[i]
        res = Particle(
            i,
            config["J"],
            config["Par"],
            mass=config["m0"],
            width=config["g0"],
        )
        chain = config["Chain"]
        if chain < 0:
            dec1 = HelicityDecay(a, [res, c])
            dec2 = HelicityDecay(res, [b, d])
        elif chain < 100:
            dec1 = HelicityDecay(a, [res, d])
            dec2 = HelicityDecay(res, [b, c])
        elif chain < 200:
            dec1 = HelicityDecay(a, [res, b])
            dec2 = HelicityDecay(res, [d, c])
        else:
            raise Exception("unknown chain")
        decay[i] = [dec1, dec2]

    decs = DecayGroup(a.chain_decay())
    return decs, [d, b, c], decay


def get_amplitude(decs, config_list, decay, polar=True):
    amp = AmplitudeModel(decs, polar)
    for i in config_list:
        if "coef_head" in config_list[i]:
            coef_head = config_list[i]["coef_head"]
            for a, b in zip(decay[i], decay[coef_head]):
                b.g_ls.sameas(a.g_ls)
            for j in decs.chains:
                if decay[i][0] in j:
                    for k in decs.chains:
                        if decay[coef_head][0] in k:
                            k.total.r_shareto(j.total)
        if "total" in config_list[i]:
            for j in decs.chains:
                if decay[i][0] in j:
                    j.total.fixed(config_list[i]["total"])
        """if ("float" in config_list[i]) and config_list[i]["float"]:
            if 'm' in config_list[i]["float"]:
                decay[i][1].core.mass.freed()
            if 'g' in config_list[i]["float"]:
                decay[i][1].core.width.freed()"""
    return amp


def load_params(amp, init_params):
    try:
        with open(init_params) as f:
            param_0 = json.load(f)
            if "value" in param_0:
                param_0 = param_0["value"]
            param = {}
            for k, v in param_0.items():
                if k.endswith(":0"):
                    k = k[:-2]
                param[k] = v
            print("using {}".format(init_params))
            amp.set_params(param)
        RDM_INI = False
    except Exception as e:
        print(e)
        RDM_INI = True
        print("using RANDOM parameters")


def fit(method="BFGS", init_params="init_params.json", hesse=True, frac=True):
    POLAR = True  # fit in polar coordinates. should be consistent with init_params.json if any
    GEN_TOY = False  # use toy data (mcdata and bg stay the same). REMEMBER to update gen_params.json

    dtype = "float64"
    w_bkg = 0.768331
    # set_gpu_mem_growth()
    tf.keras.backend.set_floatx(dtype)
    # open Resonances list as dict
    config_list = load_config_file("Resonances")

    decs, final_particles, decay = get_decay_chains(config_list)
    data, bg, mcdata = prepare_data(
        decs, particles=final_particles, dtype=dtype
    )
    amp = get_amplitude(decs, config_list, decay, polar=POLAR)
    load_params(amp, init_params)
    amp.vm.trans_params(POLAR)

    model = Model(amp, w_bkg=w_bkg)
    pprint(model.get_params())
    # print(model.Amp(data))
    # tf.summary.trace_on(graph=True, profiler = True)
    now = time.time()
    nll = model.nll(data, mcdata, bg=bg)
    print(nll)
    print(time.time() - now)
    # tf.summary.trace_export(name="sum_amp", step=0, profiler_outdir=log_dir)
    now = time.time()
    print(model.nll_grad(data, mcdata, bg=bg, batch=15000))
    print(time.time() - now)
    # now = time.time()
    # print(model.nll_grad_hessian(data, mcdata, batch=10000))
    # print(time.time() - now)
    fcn = FCN(model, data, mcdata, bg=bg, batch=65000)

    # fit configure
    args = {}
    args_name = model.Amp.vm.trainable_vars
    x0 = []
    bnds = []
    bounds_dict = {
        # "Zc_4160_m:0":(4.1,4.22),
        # "Zc_4160_g:0":(0,None)
    }

    for i in model.Amp.trainable_variables:
        args[i.name] = i.numpy()
        x0.append(i.numpy())
        if i.name in bounds_dict:
            bnds.append(bounds_dict[i.name])
        else:
            bnds.append((None, None))
        args["error_" + i.name] = 0.1

    check_grad = False
    if check_grad:
        _, gs0 = fcn.nll_grad(x0)
        gs = []
        for i, name in enumerate(args_name):
            x0[i] += 1e-5
            nll0, _ = fcn.nll_grad(x0)
            x0[i] -= 2e-5
            nll1, _ = fcn.nll_grad(x0)
            x0[i] += 1e-5
            gs.append((nll0 - nll1) / 2e-5)
            print(args_name[i], gs[i], gs0[i])

    check_hessian = False
    if check_hessian:
        nll, g, hs0 = fcn.nll_grad_hessian(x0, 20000)
        hs = []
        for i, name in enumerate(args_name):
            x0[i] += 1e-5
            _, gi1 = fcn.nll_grad(x0)
            x0[i] -= 2e-5
            _, gi2 = fcn.nll_grad(x0)
            x0[i] += 1e-5
            hs.append((gi1 - gi2) / 2e-5)
            print(args_name[i], hs[i], hs0[i])

    points = []
    nlls = []
    now = time.time()
    maxiter = 1000

    # s = basinhopping(f.nll_grad,np.array(x0),niter=6,disp=True,minimizer_kwargs={"jac":True,"options":{"disp":True}})
    if method in ["BFGS", "CG", "Nelder-Mead"]:

        def callback(x):
            if np.fabs(x).sum() > 1e7:
                x_p = dict(zip(args_name, x))
                raise Exception("x too large: {}".format(x_p))
            points.append(model.Amp.vm.get_all_val())
            nlls.append(float(fcn.cached_nll))
            # if len(nlls) > maxiter:
            #    with open("fit_curve.json", "w") as f:
            #        json.dump({"points": points, "nlls": nlls}, f, indent=2)
            #    pass  # raise Exception("Reached the largest iterations: {}".format(maxiter))
            print(fcn.cached_nll)

        # bd = Bounds(bnds)
        fcn.model.Amp.vm.set_bound(bounds_dict)
        f_g = fcn.model.Amp.vm.trans_fcn_grad(fcn.nll_grad)
        s = minimize(
            f_g,
            np.array(fcn.model.Amp.vm.get_all_val(True)),
            method=method,
            jac=True,
            callback=callback,
            options={"disp": 1, "gtol": 1e-4, "maxiter": maxiter},
        )
        xn = fcn.model.Amp.vm.get_all_val()  # bd.get_y(s.x)
    elif method in ["L-BFGS-B"]:

        def callback(x):
            if np.fabs(x).sum() > 1e7:
                x_p = dict(zip(args_name, x))
                raise Exception("x too large: {}".format(x_p))
            points.append([float(i) for i in x])
            nlls.append(float(fcn.cached_nll))

        s = minimize(
            fcn.nll_grad,
            np.array(x0),
            method=method,
            jac=True,
            bounds=bnds,
            callback=callback,
            options={
                "disp": 1,
                "maxcor": 10000,
                "ftol": 1e-15,
                "maxiter": maxiter,
            },
        )
        xn = s.x
    else:
        pass  # raise Exception("unknown method")
    print("########## fit state:")
    print(s)
    # print(s.success,s.message)
    print("\nTime for fitting:", time.time() - now)
    model.Amp.vm.trans_params(POLAR)
    val = {k: v.numpy() for k, v in model.Amp.variables.items()}
    with open("final_params.json", "w") as f:
        json.dump({"value": val}, f, indent=2)
    err = {}
    if hesse:
        inv_he = cal_hesse_error(
            model.Amp, val, w_bkg, data, mcdata, bg, args_name, batch=20000
        )
        diag_he = inv_he.diagonal()
        hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
        print(hesse_error)
        err = dict(zip(model.Amp.vm.trainable_vars, hesse_error))
    print("\n########## fit results:")
    for i in val:
        print("  ", i, ":", error_print(val[i], err.get(i, None)))

    outdic = {"value": val, "error": err, "config": config_list}
    with open("final_params.json", "w") as f:
        json.dump(outdic, f, indent=2)
    # print("\n########## ratios of partial wave amplitude square")
    # calPWratio(params,POLAR)

    if frac:
        err_frac = {}
        if hesse:
            frac, grad = cal_fitfractions(
                model.Amp, list(split_generator(mcdata, 25000))
            )
        else:
            frac = cal_fitfractions_no_grad(
                model.Amp, list(split_generator(mcdata, 45000))
            )

        for i in frac:
            if hesse:
                err_frac[i] = np.sqrt(np.dot(np.dot(inv_he, grad[i]), grad[i]))
        print("########## fit fractions")
        for i in frac:
            print(i, ":", error_print(frac[i], err_frac.get(i, None)))
    print("\nEND\n")
    """outdic = {"value": val, "error": err, "frac": frac, "err_frac": err_frac}
    with open("glbmin_params.json", "w") as f:
        json.dump(outdic, f, indent=2)
    """


def main():
    import argparse

    parser = argparse.ArgumentParser(description="simple fit scripts")
    parser.add_argument(
        "--no-hesse", action="store_false", default=True, dest="hesse"
    )
    parser.add_argument(
        "--no-frac", action="store_false", default=True, dest="frac"
    )
    parser.add_argument(
        "--no-GPU", action="store_false", default=True, dest="has_gpu"
    )
    parser.add_argument("--method", default="BFGS", dest="method")
    results = parser.parse_args()
    if results.has_gpu:
        with tf.device("/device:GPU:0"):
            fit(method=results.method, hesse=results.hesse, frac=results.frac)
    else:
        with tf.device("/device:CPU:0"):
            fit(method=results.method, hesse=results.hesse, frac=results.frac)


if __name__ == "__main__":

    # summary_writer = tf.summary.create_file_writer(log_dir)
    # with summary_writer.as_default():
    main()



import sympy as sym
import tensorflow as tf

from tf_pwa.amp import Particle
from tf_pwa.amp import get_relative_p as get_relative_p2
from tf_pwa.amp import register_particle
from tf_pwa.breit_wigner import get_bprime_coeff


def get_relative_p(m0, m1, m2):
    a = m0 ** 2 - (m1 + m2) ** 2
    b = m0 ** 2 - (m1 - m2) ** 2
    return sym.sqrt(a * b) / m0 / sym.S(2)


def Fb(l, q, d=3):
    if l == 0:
        return sym.S(1)
    z = (q * d) ** 2
    coef = get_bprime_coeff(l)
    frac = sum(coef)
    d = 0
    for i, k in enumerate(coef):
        d = d + k * z ** (l - i)
    return sym.sqrt(frac * z ** l / d)


def Bl(l, q, q0, d=3):
    return Fb(l, q, d) / Fb(l, q0, d)


def KMatrix_single(n_pole, m1, m2, l, d=3, bkg=0, Kb=0):
    m = sym.Symbol("m", positive=True)
    mi = [sym.Symbol(f"m{i}", positive=True) for i in range(n_pole)]
    g0i = [sym.Symbol(f"g{i}", real=True) for i in range(n_pole)]
    betai = [
        sym.Symbol(f"alpha{i}", real=True)
        + sym.I * sym.Symbol(f"beta{i}", real=True)
        for i in range(n_pole)
    ]
    p = sym.Symbol(f"p", positive=True)  # get_relative_p(m, m1, m2) #
    p0 = [
        sym.Symbol(f"p0{i}", positive=True) for i in range(n_pole)
    ]  # [get_relative_p(m, m1, m2) for m in mi] #
    rho = sym.S(2) * p / m
    K = 0
    for i in range(n_pole):
        tmp = (
            mi[i]
            * g0i[i]
            * (mi[i] / m)
            * (p / p0[i])
            * Bl(l, p, p0[i], d) ** 2
        )
        denominator = mi[i] ** 2 - m ** 2
        tmp = tmp / denominator
        K = tmp + K
    k = K + Kb

    P = 0
    for i in range(n_pole):
        # remove Bl function in gls
        tmp = betai[i] * mi[i] * g0i[i]  # * Bl(l, p, p0[i], d)
        P = tmp / (mi[i] * mi[i] - m * m) + P
    P = P + bkg  #  + sym.Symbol("d", complex=True)
    return P / (sym.S(1) - sym.I * K)


@register_particle("KMatrixSingleChannel")
class KmatrixSingleChannelParticle(Particle):
    """
    K matrix model for single channel multi pole.

    .. math::
        K = \\sum_{i} \\frac{m_i \\Gamma_i(m)}{m_i^2 - m^2 }

    .. math::
        P = \\sum_{i} \\frac{\\beta_i m_0 \\Gamma_0 }{ m_i^2 - m^2}

    the barrier factor is included in gls

    .. math::
        R(m) = (1-iK)^{-1} P

    """

    def __init__(self, *args, **kwargs):
        self.d = 3
        super().__init__(*args, **kwargs)
        self.n_pole = len(self.mass_list)

    def init_params(self):
        self.mi = []
        self.gi = []
        self.beta = []

        assert self.n_pole > 0
        if self.bw_l is None:
            decay = self.decay[0]
            self.bw_l = min(decay.get_l_list())

        self.symbol = sym.together(
            KMatrix_single(self.n_pole, self.m1, self.m2, self.bw_l, self.d)
        )
        self.function = opt_lambdify(
            self.symbol.free_symbols,
            (sym.re(self.symbol), sym.im(self.symbol)),
            modules="tensorflow",
        )

        for i in range(self.n_pole):
            self.mi.append(
                self.add_var(f"mass{i+1}", value=self.mass_list[i], fix=True)
            )
            self.gi.append(
                self.add_var(f"width{i+1}", value=self.width_list[i], fix=True)
            )
            self.beta.append(self.add_var(f"beta{i+1}", is_complex=True))
        self.beta[0].fixed(1.0)

    def get_mi(self):
        return [self.mi[i]() for i in range(self.n_pole)]

    def get_gi(self):
        return [self.gi[i]() for i in range(self.n_pole)]

    def get_beta(self):
        ret_r = [tf.math.real(self.beta[i]()) for i in range(self.n_pole)]
        ret_i = [tf.math.imag(self.beta[i]()) for i in range(self.n_pole)]
        return ret_r, ret_i

    def get_amp(self, *args, **kwargs):
        m = args[0]["m"]
        mi = self.get_mi()
        gi = self.get_gi()
        beta_r, beta_i = self.get_beta()
        params = {"m": m}
        params["p"] = get_relative_p2(m, self.m1, self.m2)
        for i in range(self.n_pole):
            params[f"m{i}"] = mi[i]
            params[f"g{i}"] = gi[i]
            params[f"alpha{i}"] = beta_r[i]
            params[f"beta{i}"] = beta_i[i]
            params[f"p0{i}"] = get_relative_p2(mi[i], self.m1, self.m2)
        ret_r, ret_i = self.function(**params)
        return tf.complex(ret_r, ret_i)


def opt_lambdify(args, expr, **kwargs):
    subs, ret = sym.cse(expr)
    # print(subs, ret)
    head = "def _generate_fun({}):\n".format(",".join(str(i) for i in args))
    for a, b in subs:
        head += "  {} = {}\n".format(a, b).replace("sqrt", "tf.sqrt")
    head += "  return {}\n".format(ret)
    var = {}
    exec(head, globals(), var)
    return var["_generate_fun"]



import argparse

import matplotlib.pyplot as plt
import numpy as np
import uproot
from scipy import interpolate


def mass(x):
    return x[:, 0] ** 2 - x[:, 1] ** 2 - x[:, 2] ** 2 - x[:, 3] ** 2


def get_histogram_function(file_name, branch):
    """get histogram function from root file in branch"""

    def fill_bound(x):
        x[0] = x[1] - (x[2] - x[1])
        x[-1] = x[-2] + (x[-2] - x[-3])
        return x

    # "data/EffMap_B0toD0Dspi_Run2.root"
    with uproot.open(file_name) as f:
        bg = f.get(branch)  # "RegDalitzEfficiency")

        counts, edges = bg.allnumpy()
    x, y = edges[0]  # D0barpi, Dspi
    x = fill_bound(x)
    y = fill_bound(y)
    x = (x[:-1] + x[1:]) / 2
    y = (y[:-1] + y[1:]) / 2
    f = interpolate.RectBivariateSpline(x, y, counts)
    return f, x, y


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input_file", default="data/phsp.dat")
    parser.add_argument("-o", "--out_file", default=None)
    parser.add_argument("-r", "--root_file", default="data/BG_Hist.root")
    parser.add_argument(
        "-b", "--branch_name", default="BackgroundDistribution"
    )

    args = parser.parse_args()

    p1, p2, p3 = (
        np.loadtxt(args.input_file).reshape((-1, 3, 4)).transpose((1, 0, 2))
    )

    print("m0: ", np.sqrt(mass(p1 + p2 + p3)))
    print("m1: ", np.sqrt(mass(p1)))
    print("m2: ", np.sqrt(mass(p2)))
    print("m3: ", np.sqrt(mass(p3)))

    m13 = mass(p1 + p3)
    m23 = mass(p2 + p3)

    f, x, y = get_histogram_function(args.root_file, args.branch_name)
    xx, yy = np.meshgrid(x, y)

    if args.out_file is None:
        out_file = "{}_value.dat".format(args.input_file[:-4])
    else:
        out_file = args.out_file
    np.savetxt(out_file, f.ev(m13, m23))

    plt.contourf(xx, yy, f(x, y), vmin=1e-6)
    plt.colorbar()
    plt.xlabel("$M^2_{13}$")
    plt.ylabel("$M^2_{23}$")
    plt.savefig("bg_histogram.png")


if __name__ == "__main__":
    main()



import tensorflow as tf


def nll_funciton(f, data, phsp):
    """
    nagtive log likelihood for minimize

    """

    def g():
        ll = tf.math.log(f(data))
        int_mc = tf.reduce_mean(f(phsp))
        return -tf.reduce_sum(ll) + ll.shape[0] * tf.math.log(int_mc)

    return g



import os.path
import sys

import yaml

# import tf_pwa
from tf_pwa.config_loader import ConfigLoader

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


def main():
    """Calculate errors of a given set of parameters and their correlation coefficients."""
    import argparse

    parser = argparse.ArgumentParser(
        description="calculate errors of a given set of parameters and their correlation coefficients"
    )
    parser.add_argument(
        "--params", default="final_params.json", dest="params_file"
    )
    results = parser.parse_args()

    plot(params_file=results.params_file)


def plot(params_file):
    config = ConfigLoader("config.yml")
    with open(params_file) as f:
        params = yaml.safe_load(f)
    params = params["value"]
    config.plot_partial_wave(
        params, plot_pull=True, prefix="fig/", save_pdf=True, save_root=True
    )


if __name__ == "__main__":
    main()



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

import os
import shutil
import subprocess
import sys

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")

from tf_pwa.amp import PARTICLE_MODEL, get_config
from tf_pwa.experimental import (  # type: ignore  # pylint: disable=unused-import
    extra_amp,
)

# -- Project information -----------------------------------------------------
project = "TFPWA"
copyright = "2020, Yi Jiang"  # pylint: disable=redefined-builtin
author = "Yi Jiang"

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_gallery.gen_gallery",
]
exclude_patterns = [
    ".DS_Store",
    "Thumbs.db",
    "_build",
]
source_suffix = [
    ".rst",
]

# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_rtd_theme"
html_title = "TFPWA"
viewcode_follow_imported_members = True

# -- Options for API ---------------------------------------------------------
add_module_names = False
autodoc_mock_imports = [
    "iminuit",
    "tensorflow",
]

# Cross-referencing configuration
default_role = "py:obj"
primary_domain = "py"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# -- Generate API skeleton ----------------------------------------------------
shutil.rmtree("api", ignore_errors=True)
subprocess.call(
    " ".join(
        [
            "sphinx-apidoc",
            "-o api/",
            "--force",
            "--no-toc",
            "--templatedir _templates",
            "--separate",
            "../tf_pwa/",
            # exclude patterns
            "../tf_pwa/tests",
            "../tf_pwa/config_loader/tests/*",
        ]
    ),
    shell=True,
)


# -- Generate available resonance models --------------------------------------
def add_indent(s, number=2):
    ret = ""
    for i in s.split("\n"):
        ret += " " * number + i + "\n"
    return ret


def gen_particle_model():
    particle_model_doc = """
--------------------------
Available Resonances Model
--------------------------

"""
    models = []
    model_params = {}
    for idx, (k, v) in enumerate(get_config(PARTICLE_MODEL).items(), 1):
        doc_i = v.__doc__
        if v.__doc__ is None and v.get_amp.__doc__ is None:
            continue
        if v.__doc__ is None:
            doc_i = v.get_amp.__doc__

        if v not in models:
            models.append(v)
        if v in model_params:
            model_params[v]["name"].append(f'"{k}"')
        else:
            model_params[v] = {"name": [f'"{k}"'], "doc": doc_i}

    for idx, v in enumerate(models):
        name_list = model_params[v]["name"]
        name = ", ".join(name_list)
        doc_i = model_params[v]["doc"]
        particle_model_doc += (
            f"\n{idx+1}. :code:`{name}`"
            f" (`~{v.__module__}.{v.__qualname__}`)\n\n"
        )
        idx += 1
        particle_model_doc += add_indent(doc_i) + "\n\n"

    with open(
        os.path.dirname(os.path.abspath(__file__)) + "/particle_model.rst", "w"
    ) as f:
        f.write(particle_model_doc)


gen_particle_model()


sphinx_gallery_conf = {
    "examples_dirs": "../examples",  # path to your example scripts
    "gallery_dirs": "auto_examples",  # path to where to save gallery generated output
    "line_numbers": True,
    "run_stale_examples": True,
    "filename_pattern": "ex",
}



import contextlib
import copy
import functools
import itertools
import json
import os
import re
import time
import warnings

import matplotlib.pyplot as plt
import numpy as np
import sympy as sy
import yaml
from scipy.interpolate import interp1d
from scipy.optimize import BFGS, basinhopping, minimize

from tf_pwa.amp import (
    AmplitudeModel,
    DecayChain,
    DecayGroup,
    get_decay,
    get_particle,
)
from tf_pwa.applications import (
    cal_hesse_error,
    corr_coef_matrix,
    fit,
    fit_fractions,
)
from tf_pwa.cal_angle import prepare_data_from_decay
from tf_pwa.data import (
    data_index,
    data_shape,
    data_split,
    load_data,
    save_data,
)
from tf_pwa.fit import FitResult
from tf_pwa.fit_improve import minimize as my_minimize
from tf_pwa.model import FCN, CombineFCN, Model, Model_new
from tf_pwa.model.cfit import Model_cfit
from tf_pwa.particle import split_particle_type
from tf_pwa.root_io import has_uproot, save_dict_to_root
from tf_pwa.utils import time_print
from tf_pwa.variable import Variable, VarsManager

from .config_loader import ConfigLoader
from .decay_config import DecayConfig


class MultiConfig(object):
    def __init__(self, file_names, vm=None, total_same=False, share_dict={}):
        if vm is None:
            self.vm = VarsManager()
            # print(self.vm)
        else:
            self.vm = vm
        self.total_same = total_same
        self.configs = [
            ConfigLoader(i, vm=self.vm, share_dict=share_dict)
            for i in file_names
        ]
        self.bound_dic = {}
        self.gauss_constr_dic = {}
        self._neglect_when_set_params = []
        self.cached_fcn = {}
        self.inv_he = None

    def get_all_data(self):
        return [i.get_all_data() for i in self.configs]

    def get_amplitudes(self, vm=None):
        if not self.total_same:
            amps = [
                j.get_amplitude(name="s" + str(i), vm=vm)
                for i, j in enumerate(self.configs)
            ]
        else:
            amps = [j.get_amplitude(vm=vm) for j in self.configs]
        for i in self.configs:
            self.bound_dic.update(i.bound_dic)
            self.gauss_constr_dic.update(i.gauss_constr_dic)
            for j in i._neglect_when_set_params:
                if j not in self._neglect_when_set_params:
                    self._neglect_when_set_params.append(j)
        return amps

    """
    def _get_models(self, vm=None): # get_model is useless to users given get_fcn and get_amplitude
        if not self.total_same:
            models = [j._get_model(name="s"+str(i), vm=vm)
                      for i, j in enumerate(self.configs)]
        else:
            models = [j._get_model(vm=vm) for j in self.configs]
        return models
    """

    def get_fcns(self, datas=None, vm=None, batch=65000):
        if datas is not None:
            if not self.total_same:
                fcns = [
                    i[1].get_fcn(
                        name="s" + str(i[0]), all_data=j, vm=vm, batch=batch
                    )
                    for i, j in zip(enumerate(self.configs), datas)
                ]
            else:
                fcns = [
                    j.get_fcn(all_data=data, vm=vm, batch=batch)
                    for data, j in zip(datas, self.configs)
                ]
        else:
            if not self.total_same:
                fcns = [
                    j.get_fcn(name="s" + str(i), vm=vm, batch=batch)
                    for i, j in enumerate(self.configs)
                ]
            else:
                fcns = [j.get_fcn(vm=vm, batch=batch) for j in self.configs]
        return fcns

    def get_fcn(self, datas=None, vm=None, batch=65000):
        if datas is None:
            if vm in self.cached_fcn:
                return self.cached_fcn[vm]
        fcns = self.get_fcns(datas=datas, vm=vm, batch=batch)
        fcn = CombineFCN(fcns=fcns, gauss_constr=self.gauss_constr_dic)
        if datas is None:
            self.cached_fcn[vm] = fcn
        return fcn

    def get_args_value(self, bounds_dict):
        args = {}
        args_name = self.vm.trainable_vars
        x0 = []
        bnds = []

        for i in self.vm.trainable_variables:
            args[i.name] = i.numpy()
            x0.append(i.numpy())
            if i.name in bounds_dict:
                bnds.append(bounds_dict[i.name])
            else:
                bnds.append((None, None))
            args["error_" + i.name] = 0.1

        return args_name, x0, args, bnds

    @time_print
    def fit(self, datas=None, batch=65000, method="BFGS", maxiter=None):
        fcn = self.get_fcn(datas=datas)
        # fcn.gauss_constr.update({"Zc_Xm_width": (0.177, 0.03180001857)})
        print("\n########### initial parameters")
        print(json.dumps(fcn.get_params(), indent=2), flush=True)
        print("initial NLL: ", fcn({}))
        self.fit_params = fit(
            fcn=fcn, method=method, bounds_dict=self.bound_dic, maxiter=maxiter
        )
        if self.fit_params.hess_inv is not None:
            self.inv_he = self.fit_params.hess_inv
        """# fit configure
        bounds_dict = {}
        args_name, x0, args, bnds = self.get_args_value(bounds_dict)

        points = []
        nlls = []
        now = time.time()
        maxiter = 1000
        min_nll = 0.0
        ndf = 0

        if method in ["BFGS", "CG", "Nelder-Mead"]:
            def callback(x):
                if np.fabs(x).sum() > 1e7:
                    x_p = dict(zip(args_name, x))
                    raise Exception("x too large: {}".format(x_p))
                points.append(self.vm.get_all_val())
                nlls.append(float(fcn.cached_nll))
                # if len(nlls) > maxiter:
                #    with open("fit_curve.json", "w") as f:
                #        json.dump({"points": points, "nlls": nlls}, f, indent=2)
                #    pass  # raise Exception("Reached the largest iterations: {}".format(maxiter))
                print(fcn.cached_nll)

            self.vm.set_bound(bounds_dict)
            f_g = self.vm.trans_fcn_grad(fcn.nll_grad)
            s = minimize(f_g, np.array(self.vm.get_all_val(True)), method=method,
                         jac=True, callback=callback, options={"disp": 1, "gtol": 1e-4, "maxiter": maxiter})
            xn = s.x  # self.vm.get_all_val()  # bd.get_y(s.x)
            ndf = s.x.shape[0]
            min_nll = s.fun
            if hasattr(s, "hess_inv"):
                self.inv_he = s.hess_inv
            success = s.success
        elif method in ["L-BFGS-B"]:
            def callback(x):
                if np.fabs(x).sum() > 1e7:
                    x_p = dict(zip(args_name, x))
                    raise Exception("x too large: {}".format(x_p))
                points.append([float(i) for i in x])
                nlls.append(float(fcn.cached_nll))

            s = minimize(fcn.nll_grad, np.array(x0), method=method, jac=True, bounds=bnds, callback=callback,
                         options={"disp": 1, "maxcor": 10000, "ftol": 1e-15, "maxiter": maxiter})
            xn = s.x
            ndf = s.x.shape[0]
            min_nll = s.fun
            success = s.success
        elif method in ["iminuit"]:
            from .fit import fit_minuit
            m = fit_minuit(fcn)
            return m
        else:
            raise Exception("unknown method")
        self.vm.set_all(xn)
        params = self.vm.get_all_dic()
        return FitResult(params, fcn, min_nll, ndf=ndf, success=success)"""
        return self.fit_params

    def reinit_params(self):
        self.get_fcn().vm.refresh_vars(self.bound_dic)

    def get_params_error(
        self, params=None, datas=None, batch=10000, using_cached=False
    ):
        if params is None:
            params = {}
        if hasattr(params, "params"):
            params = getattr(params, "params")
        fcn = self.get_fcn(datas, batch=batch)
        if using_cached and self.inv_he is not None:
            hesse_error = np.sqrt(np.fabs(self.inv_he.diagonal())).tolist()
        else:
            hesse_error, self.inv_he = cal_hesse_error(
                fcn, params, check_posi_def=True, save_npy=True
            )
        # hesse_error, self.inv_he = cal_hesse_error(
        # fcn, params, check_posi_def=True, save_npy=True
        # )
        print("hesse_error:", hesse_error)
        err = dict(zip(self.vm.trainable_vars, hesse_error))
        if hasattr(self, "fit_params"):
            self.fit_params.set_error(err)
        return err

    def get_params(self, trainable_only=False):
        # _amps = self.get_fcn()
        return self.vm.get_all_dic(trainable_only)

    def set_params(self, params, neglect_params=None):
        _amps = self.get_amplitudes()
        if isinstance(params, str):
            if params == "":
                return False
            try:
                with open(params) as f:
                    params = yaml.safe_load(f)
            except Exception as e:
                print(e)
                return False
        if hasattr(params, "params"):
            params = params.params
        if isinstance(params, dict):
            if "value" in params:
                params = params["value"]
        ret = params.copy()
        if neglect_params is None:
            neglect_params = self._neglect_when_set_params
        if len(neglect_params) != 0:
            warnings.warn(
                "Neglect {} when setting params.".format(neglect_params)
            )
            for v in params:
                if v in self._neglect_when_set_params:
                    del ret[v]
        self.vm.set_all(ret)
        return True

    @contextlib.contextmanager
    def params_trans(self):
        with self.vm.error_trans(self.inv_he) as f:
            yield f

    def save_params(self, file_name):
        params = self.get_params()
        val = {k: float(v) for k, v in params.items()}
        with open(file_name, "w") as f:
            json.dump(val, f, indent=2)



#!/usr/bin/env python3

import csv
import json
import time
from pprint import pprint

# avoid using Xwindow
import matplotlib

matplotlib.use("agg")

import tensorflow as tf

# examples of custom particle model
from tf_pwa.amp import simple_resonance
from tf_pwa.config_loader import ConfigLoader, MultiConfig
from tf_pwa.experimental import extra_amp, extra_data
from tf_pwa.utils import error_print, tuple_table


@simple_resonance("New", params=["alpha", "beta"])
def New_Particle(m, alpha, beta=0):
    """example Particle model define, can be used in config.yml as `model: New`"""
    zeros = tf.zeros_like(m)
    r = -tf.complex(alpha, beta) * tf.complex(m, zeros)
    return tf.exp(r)


def json_print(dic):
    """print parameters as json"""
    s = json.dumps(dic, indent=2)
    print(s, flush=True)


def load_config(config_file="config.yml", total_same=False):
    config_files = config_file.split(",")
    if len(config_files) == 1:
        return ConfigLoader(config_files[0])
    return MultiConfig(config_files, total_same=total_same)


def fit(config, init_params="", method="BFGS", loop=1, maxiter=500):
    """
    simple fit script
    """
    # load config.yml
    # config = ConfigLoader(config_file)

    # load data
    all_data = config.get_all_data()

    fit_results = []
    for i in range(loop):
        # set initial parameters if have
        if config.set_params(init_params):
            print("using {}".format(init_params))
        else:
            print("\nusing RANDOM parameters", flush=True)
        # try to fit
        try:
            fit_result = config.fit(
                batch=65000, method=method, maxiter=maxiter
            )
        except KeyboardInterrupt:
            config.save_params("break_params.json")
            raise
        except Exception as e:
            print(e)
            config.save_params("break_params.json")
            raise
        fit_results.append(fit_result)
        # reset parameters
        try:
            config.reinit_params()
        except Exception as e:
            print(e)

    fit_result = fit_results.pop()
    for i in fit_results:
        if i.success:
            if not fit_result.success or fit_result.min_nll > i.min_nll:
                fit_result = i

    config.set_params(fit_result.params)
    json_print(fit_result.params)
    fit_result.save_as("final_params.json")

    # calculate parameters error
    if maxiter != 0:
        fit_error = config.get_params_error(fit_result, batch=13000)
        fit_result.set_error(fit_error)
        fit_result.save_as("final_params.json")
        pprint(fit_error)

        print("\n########## fit results:")
        print("Fit status: ", fit_result.success)
        print("Minimal -lnL = ", fit_result.min_nll)
        for k, v in config.get_params().items():
            print(k, error_print(v, fit_error.get(k, None)))

    return fit_result


def write_some_results(config, fit_result, save_root=False):
    # plot partial wave distribution
    config.plot_partial_wave(fit_result, plot_pull=True, save_root=save_root)

    # calculate fit fractions
    phsp_noeff = config.get_phsp_noeff()
    fit_frac, err_frac = config.cal_fitfractions({}, phsp_noeff)

    print("########## fit fractions")
    fit_frac_string = ""
    for i in fit_frac:
        if isinstance(i, tuple):
            name = "{}x{}".format(*i)
        else:
            name = i
        fit_frac_string += "{} {}\n".format(
            name, error_print(fit_frac[i], err_frac.get(i, None))
        )
    print(fit_frac_string)
    save_frac_csv("fit_frac.csv", fit_frac)
    save_frac_csv("fit_frac_err.csv", err_frac)
    # from frac_table import frac_table
    # frac_table(fit_frac_string)
    # chi2, ndf = config.cal_chi2(mass=["R_BC", "R_CD"], bins=[[2,2]]*4)


def write_some_results_combine(config, fit_result, save_root=False):

    from tf_pwa.applications import fit_fractions

    for i, c in enumerate(config.configs):
        c.plot_partial_wave(
            fit_result, prefix="figure/s{}_".format(i), save_root=save_root
        )

    for it, config_i in enumerate(config.configs):
        print("########## fit fractions {}:".format(it))
        mcdata = config_i.get_phsp_noeff()
        fit_frac, err_frac = fit_fractions(
            config_i.get_amplitude(),
            mcdata,
            config.inv_he,
            fit_result.params,
        )
        fit_frac_string = ""
        for i in fit_frac:
            if isinstance(i, tuple):
                name = "{}x{}".format(*i)  # interference term
            else:
                name = i  # fit fraction
            fit_frac_string += "{} {}\n".format(
                name, error_print(fit_frac[i], err_frac.get(i, None))
            )
        print(fit_frac_string)
        save_frac_csv(f"fit_frac{it}.csv", fit_frac)
        save_frac_csv(f"fit_frac{it}_err.csv", err_frac)
    # from frac_table import frac_table
    # frac_table(fit_frac_string)


def save_frac_csv(file_name, fit_frac):
    table = tuple_table(fit_frac)
    with open(file_name, "w") as f:
        f_csv = csv.writer(f)
        f_csv.writerows(table)


def write_run_point():
    """ write time as a point of fit start"""
    with open(".run_start", "w") as f:
        localtime = time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(time.time())
        )
        f.write(localtime)


def main():
    """entry point of fit. add some arguments in commond line"""
    import argparse

    parser = argparse.ArgumentParser(description="simple fit scripts")
    parser.add_argument(
        "--no-GPU", action="store_false", default=True, dest="has_gpu"
    )
    parser.add_argument("-c", "--config", default="config.yml", dest="config")
    parser.add_argument(
        "-i", "--init_params", default="init_params.json", dest="init"
    )
    parser.add_argument("-m", "--method", default="BFGS", dest="method")
    parser.add_argument("-l", "--loop", type=int, default=1, dest="loop")
    parser.add_argument(
        "-x", "--maxiter", type=int, default=2000, dest="maxiter"
    )
    parser.add_argument("-r", "--save_root", default=False, dest="save_root")
    parser.add_argument(
        "--total-same", action="store_true", default=False, dest="total_same"
    )
    results = parser.parse_args()
    if results.has_gpu:
        devices = "/device:GPU:0"
    else:
        devices = "/device:CPU:0"
    with tf.device(devices):
        config = load_config(results.config, results.total_same)
        fit_result = fit(
            config, results.init, results.method, results.loop, results.maxiter
        )
        if isinstance(config, ConfigLoader):
            write_some_results(config, fit_result, save_root=results.save_root)
        else:
            write_some_results_combine(
                config, fit_result, save_root=results.save_root
            )


if __name__ == "__main__":
    write_run_point()
    main()



import tensorflow as tf

from tf_pwa.data import data_shape
from tf_pwa.utils import time_print

from .opt_int import split_gls


def build_sum_amplitude(dg, dec_chain, data):
    cached = []
    for i, dc in split_gls(dec_chain):
        amp = dg.get_amp(data)
        int_mc = amp
        m_dep = dg.get_m_dep(data)
        m_dep_all = 1.0
        for i in m_dep:
            for j in i:
                m_dep_all *= tf.reshape(j, (-1,))
        m_dep_all = tf.reshape(m_dep_all, [-1] + [1] * (len(amp.shape) - 1))
        cached.append(amp / m_dep_all)
    return cached


def build_amp_matrix(dec, data, weight=None):
    hij = []
    used_chains = dec.chains_idx
    index = []
    for k, i in enumerate(dec):
        dec.set_used_chains([k])
        tmp = []
        for j, amp in enumerate(build_sum_amplitude(dec, i, data)):
            tmp.append(amp)
        hij.append(tmp)
    dec.set_used_chains(used_chains)
    # print([i.shape for i in hij.values()])
    # print([[j.shape for j in i] for i in hij])
    return index, hij


def _prod(ls):
    ret = 1
    for i in ls:
        ret *= i
    return ret


def build_params_vector(dg, data):
    n_data = data_shape(data)
    m_dep = dg.get_m_dep(data)
    ret = []
    for i in m_dep:
        tmp = i[0]
        if tmp.shape[0] == 1:
            tmp = tf.tile(tmp, [n_data] + [1] * (len(tmp.shape) - 1))
        tmp = tf.reshape(tmp, (-1, _prod(tmp.shape[1:])))
        for j in i[1:]:
            tmp2 = tf.reshape(j, (-1, _prod(j.shape[1:])))
            tmp = tmp[:, :, None] * tmp2[:, None, :]
            tmp = tf.reshape(tmp, (-1, _prod(tmp.shape[1:])))
        ret.append(tmp)
    return ret


def cached_amp2s(dg, data):

    _amp = cached_amp(dg, data)

    @time_print
    @tf.function
    def _amp2s():  # pragma: no cover # because of tf_funciton
        amp = _amp()
        amp2s = tf.math.real(amp * tf.math.conj(amp))
        return tf.reduce_sum(amp2s, list(range(1, len(amp2s.shape))))

    return _amp2s


def build_amp2s(dg):
    @tf.function(experimental_relax_shapes=True)
    def _amp2s(data, cached_data):
        n_data = data_shape(data)
        pv = build_params_vector(dg, data)
        ret = []
        for i, j in zip(pv, cached_data):
            # print(j)
            # print(i.shape)
            a = tf.reshape(i, [-1, i.shape[1]] + [1] * (len(j[0].shape) - 1))
            ret.append(tf.reduce_sum(a * tf.stack(j, axis=1), axis=1))
        # print(ret)
        amp = tf.reduce_sum(ret, axis=0)
        amp2s = tf.math.real(amp * tf.math.conj(amp))
        return tf.reduce_sum(amp2s, list(range(1, len(amp2s.shape))))

    return _amp2s


def build_sum_angle_amplitude(dg, dec_chain, data):
    cached = []
    for i, dc in split_gls(dec_chain):
        amp = dg.get_angle_amp(data)
        cached.append(amp)
    return cached


def build_angle_amp_matrix(dec, data, weight=None):
    hij = []
    used_chains = dec.chains_idx
    for k, i in enumerate(dec):
        dec.set_used_chains([k])
        tmp = []
        for j, amp in enumerate(build_sum_angle_amplitude(dec, i, data)):
            tmp.append(amp)
        hij.append(tmp)
    dec.set_used_chains(used_chains)
    return list(dec), hij


def amp_matrix_as_dict(dec, hij):
    ret = {}
    for i, d in zip(hij, dec):
        ret[d] = i
    return ret


def cached_amp(dg, data, matrix_method=build_angle_amp_matrix):

    idx, c_amp = matrix_method(dg, data)
    n_data = data_shape(data)

    @tf.function
    def _amp():
        pv = build_params_vector(dg, data)
        ret = []
        for i, j in zip(pv, c_amp):
            a = tf.reshape(i, [n_data, -1] + [1] * (len(j[0].shape) - 1))
            ret.append(tf.reduce_sum(a * tf.stack(j, axis=1), axis=1))
        # print(ret)
        amp = tf.reduce_sum(ret, axis=0)
        return amp

    return _amp



import os
import time

import matplotlib
import numpy as np
import pytest
import tensorflow as tf

from tf_pwa import set_random_seed
from tf_pwa.applications import gen_data, gen_mc
from tf_pwa.config_loader import ConfigLoader, MultiConfig
from tf_pwa.experimental import build_amp
from tf_pwa.utils import tuple_table

matplotlib.use("agg")


this_dir = os.path.dirname(os.path.abspath(__file__))


def generate_phspMC(Nmc):
    """Generate PhaseSpace MC of size Nmc and save it as txt file"""
    # masses of mother particle A and daughters BCD
    mA = 4.6
    mB = 2.00698
    mC = 2.01028
    mD = 0.13957
    # a2bcd is a [3*Nmc, 4] array, which are the momenta of BCD in the rest frame of A
    a2bcd = gen_mc(mA, [mB, mC, mD], Nmc)
    return a2bcd


def generate_toy_from_phspMC(Ndata, mc_file, data_file):
    """Generate toy using PhaseSpace MC from mc_file"""
    config = ConfigLoader(f"{this_dir}/config_toy.yml")
    config.set_params(f"{this_dir}/gen_params.json")
    amp = config.get_amplitude()
    data = gen_data(
        amp,
        Ndata=Ndata,
        mcfile=mc_file,
        genfile=data_file,
        particles=config.get_dat_order(),
    )
    return data


@pytest.fixture
def gen_toy():
    set_random_seed(1)
    if not os.path.exists("toy_data"):
        os.mkdir("toy_data")
    phsp = generate_phspMC(10000)
    np.savetxt("toy_data/PHSP.dat", phsp)
    generate_toy_from_phspMC(1000, "toy_data/PHSP.dat", "toy_data/data.dat")
    bg = generate_phspMC(1000)
    data = np.loadtxt("toy_data/data.dat")
    np.savetxt("toy_data/data.dat", np.concatenate([data, bg[:300, :]]))
    np.savetxt("toy_data/bg.dat", bg)
    np.savetxt("toy_data/data_bg_value.dat", np.ones((1000 + 100,)))
    np.savetxt("toy_data/data_eff_value.dat", np.ones((1000 + 100,)))
    np.savetxt("toy_data/phsp_bg_value.dat", np.ones((10000,)))
    np.savetxt("toy_data/phsp_eff_value.dat", np.ones((10000,)))


@pytest.fixture
def toy_config(gen_toy):
    config = ConfigLoader(f"{this_dir}/config_toy.yml")
    config.set_params(f"{this_dir}/exp_params.json")
    return config


def test_build_angle_amplitude(toy_config):
    data = toy_config.get_data("data")
    dec = toy_config.get_amplitude().decay_group
    amp_dict = build_amp.build_angle_amp_matrix(dec, data[0])
    assert len(amp_dict[1]) == 3


@pytest.fixture
def toy_config2(gen_toy, fit_result):
    config = MultiConfig(
        [f"{this_dir}/config_toy.yml", f"{this_dir}/config_toy2.yml"]
    )
    config.set_params(f"{this_dir}/exp2_params.json")
    return config


@pytest.fixture
def toy_config3(gen_toy):
    config = ConfigLoader(f"{this_dir}/config_toy3.yml")
    config.set_params(f"{this_dir}/exp_params.json")
    return config


@pytest.fixture
def fit_result(toy_config):
    return toy_config.fit()


def test_save_model(toy_config):
    config = ConfigLoader(f"{this_dir}/config_toy.yml")
    config.set_params(f"{this_dir}/gen_params.json")
    config.save_tensorflow_model("toy_data/model")


def test_cfit(gen_toy):
    config = ConfigLoader(f"{this_dir}/config_cfit.yml")
    config.set_params(f"{this_dir}/gen_params.json")
    fcn = config.get_fcn()
    fcn({})
    fcn.nll_grad({})
    config.plot_partial_wave()


def test_cfit_cached(gen_toy):
    config = ConfigLoader(f"{this_dir}/config_cfit_cached.yml")
    config.set_params(f"{this_dir}/gen_params.json")
    fcn = config.get_fcn()
    fcn({})
    fcn.nll_grad({})


def test_constrains(gen_toy):
    config = ConfigLoader(f"{this_dir}/config_cfit.yml")
    var_name = "A->R_CD.B_g_ls_1r"
    config.config["constrains"]["init_params"] = {var_name: 1.0}

    @config.register_extra_constrains("init_params")
    def float_var(amp, params=None):
        amp.set_params(params)

    config.register_extra_constrains("init_params2", float_var)

    amp = config.get_amplitude()
    assert amp.get_params()[var_name] == 1.0


def test_fit(toy_config, fit_result):
    toy_config.plot_partial_wave(prefix="toy_data/figure", save_root=True)
    toy_config.plot_partial_wave(
        prefix="toy_data/figure", plot_pull=True, single_legend=True
    )
    toy_config.plot_partial_wave(
        prefix="toy_data/figure", smooth=False, bin_scale=1
    )
    toy_config.plot_partial_wave(prefix="toy_data/figure", color_first=False)
    toy_config.get_params_error(fit_result)
    toy_config.get_params_error(fit_result, method="3-point")
    toy_config.get_params_error(
        fit_result, correct_params=["A->R_CD.BR_CD->C.D_total_0i"]
    )
    fit_result.save_as("toy_data/final_params.json")
    fit_frac, frac_err = toy_config.cal_fitfractions()
    tuple_table(fit_frac)

    with toy_config.params_trans() as pt:
        a = pt["A->R_BC.D_g_ls_1r"]
        b = pt["A->R_BC.D_g_ls_1i"]
        alpha = 2 * a / (1 + a * a + b + b)
    alpha_err = pt.get_error(alpha)
    with toy_config.params_trans() as pt:
        a = pt["A->R_BC.D_g_ls_1r"]
        b = pt["A->R_BC.D_g_ls_1i"]
        x = a + b
        y = a - b
    xy_err = pt.get_error_matrix([x, y])


def test_cal_chi2(toy_config, fit_result):
    toy_config.cal_chi2(bins=[[2, 2]] * 2, mass=["R_BD", "R_CD"])


def test_cal_signal_yields(toy_config, fit_result):
    toy_config.cal_signal_yields()


def test_fit_combine(toy_config2):
    toy_config2.fit()
    toy_config2.get_params_error()
    print(toy_config2.get_params())


def test_mix_likelihood(toy_config3):
    toy_config3.fit(maxiter=1)



"""This module provides the function **cg_coef()** to calculate the Clebsch-Gordan coefficients :math:`\\langle
j_1m_1j_2m_2|JM\\rangle`.

This function has interface to `SymPy <https://www.sympy.org/en/index.html>`_ functions if it's installed correctly.
Otherwise, it will depend on the input file **tf_pwa/cg_table.json**.
"""

import json
import os

has_sympy = True
try:
    from sympy.physics.quantum.cg import CG
except ImportError:
    has_sympy = False


def cg_coef(jb, jc, mb, mc, ja, ma):
    """
    It returns the CG coefficient :math:`\\langle j_bm_bj_cm_c|j_am_a\\rangle`, as in a decay from particle *a* to *b*
    and *c*. It will either call **sympy.physics.quantum.cg()** or **get_cg_coef()**.
    """
    if has_sympy:
        return float(CG(jb, mb, jc, mc, ja, ma).doit().evalf())
    else:
        return get_cg_coef(jb, jc, mb, mc, ja, ma)


_dirname = os.path.dirname(os.path.abspath(__file__))

with open(_dirname + "/cg_table.json") as f:
    cg_table = json.load(f)


def get_cg_coef(j1, j2, m1, m2, j, m):
    """
    If SymPy is not installed, [deprecation] **cg_coef()** will call this function and derive the CG coefficient by searching into
    **tf_pwa/cg_table.json**.

    In fact, **tf_pwa/cg_table.json** only stores some of the coefficients, the others will be
    obtained by this function using some symmetry properties of the CG table.

    .. note:: **tf_pwa/cg_table.json** only contains the cases where :math:`j_1,j_2\\leqslant4`, but this should be enough for most cases in PWA.
    """
    assert m1 + m2 == m
    assert j1 >= 0
    assert j2 >= 0
    assert j >= 0
    if j1 == 0 or j2 == 0:
        return 1.0
    sign = 1
    if j1 < j2:
        if (j1 + j2 - j) % 2 == 1:
            sign = -1
        j1, j2 = j2, j1
        m1, m2 = m2, m1

    def find_cg_table(j1, j2, m1, m2, j, m):
        try:
            return cg_table[str(j1)][str(j2)][str(m1)][str(m2)][str(j)][str(m)]
        except:
            return 0.0

    return sign * find_cg_table(j1, j2, m1, m2, j, m)



import pytest

from tf_pwa.particle import *


def test_particle():
    a = BaseParticle("a", 1, -1)
    b = BaseParticle("b", 1, -1)
    c = BaseParticle("c", 0, -1)
    d = BaseParticle("d", 1, -1)
    tmp = BaseParticle("tmp", 1, -1)
    tmp2 = BaseParticle("tmp2", 1, -1)
    decay = Decay(a, [tmp, c])
    decay2 = Decay(tmp, [b, d])
    decay3 = Decay(a, [tmp2, d])
    decay4 = Decay(tmp2, [b, c])
    decaychain = DecayChain([decay, decay2])
    decaychain2 = DecayChain([decay3, decay4])
    decaygroup = DecayGroup([decaychain, decaychain2])
    print(decay.get_cg_matrix().T)
    print(np.array(decay.get_ls_list()))
    print(np.array(decay.get_ls_list())[:, 0])
    print(decaychain)
    print(decaychain.sorted_table())
    print(decaygroup)
    print(a.get_resonances())


def test_ls():
    a = BaseParticle("a", 1, -1, C=1)
    b = BaseParticle("b", 1, -1)
    c = BaseParticle("c", 1, -1)
    dec = Decay(a, [b, c])
    ls = dec.get_ls_list()
    assert len(ls) == 4
    a = BaseParticle("a", 1, -1, C=1)
    b = BaseParticle("b", 1, -1)
    c = BaseParticle("c", 1, -1)
    dec = Decay(a, [b, c], c_break=False)
    ls = dec.get_ls_list()


def test_topology():
    a = BaseParticle("a")
    b = BaseParticle("B")
    c = BaseParticle("C")
    d = BaseParticle("D")
    r = BaseParticle("R")
    d1 = Decay(a, [r, c])
    d2 = Decay(r, [b, d])
    dec = DecayChain([d1, d2])
    print(dec.standard_topology())


def test_baseparticle():
    a = BaseParticle("a")
    assert str(a) == "a"
    b = BaseParticle("a:1")
    assert str(b) == "a:1"
    assert a.name == "a"
    c = BaseParticle("a", id_=1)
    assert str(c) == "a:1"
    assert a.name == "a"
    assert b == c
    assert a < c
    assert a < "a:1"
    assert c > a
    assert c > "a"


def test_basedecay():
    a = BaseParticle("a")
    b = BaseParticle("a:1")
    c = BaseParticle("c")
    d = BaseParticle("d")
    de = BaseDecay(c, [a, b])
    assert len(c.decay) == 1
    de2 = BaseDecay(d, [a, b], disable=True)
    assert len(c.decay) == 1
    c.remove_decay(de)
    assert len(c.decay) == 0
    de2 < de
    de2 > de
    assert de2 != de
    de2 > "s"
    de < "s"
    assert de != "s"


def test_sorted_table():
    a = {
        "a": ["b", "c", "d"],
        "r": ["c", "d"],
        "b": ["b"],
        "c": ["c"],
        "d": ["d"],
    }
    de = DecayChain.from_sorted_table(a)
    print(de.sorted_table_layers())
    b = {"a": ["b", "c", "d"], "b": ["b"], "c": ["c"], "d": ["d"]}
    de2 = DecayChain.from_sorted_table(b)
    print(de2.sorted_table_layers())
    c = {"a": ["b", "c", "d"], "c": ["c"], "d": ["d"]}
    with pytest.raises(Exception):
        de3 = DecayChain.from_sorted_table(c)


def test_depth_fisrt():
    decs = DecayChain.from_particles("a", ["B", "C", "D", "E"])
    for de in decs:
        used_node = [de.top]
        for i, dec in de.depth_first():
            assert dec.core in used_node
            for j in dec.outs:
                used_node.append(j)
        used_node = [de.top]
        for i, dec in list(de.depth_first(False))[::-1]:
            assert dec.core in used_node
            for j in dec.outs:
                used_node.append(j)



import numpy as np

import tf_pwa.amp.Kmatrix
from tf_pwa.amp import get_decay, get_particle, get_relative_p


def check_frac(x, y):
    r = x / y
    assert np.allclose(r, np.mean(r))


def test_a():
    a = get_particle(
        "a",
        mass_list=[2.0],
        width_list=[0.03],
        m1=1.1,
        m2=0.5,
        bw_l=1,
        model="KMatrixSingleChannel",
    )
    b = get_particle("b", J=1, P=-1, mass=2.0, width=0.03, model="BWR")
    c = get_particle("c", J=0, P=-1, mass=1.1, width=0.03, model="one")
    d = get_particle("d", J=0, P=-1, mass=0.5, width=0.03, model="one")
    dec = get_decay(b, [c, d])
    a.init_params()
    b.init_params()

    m = np.array([1.9, 2.0, 2.1])
    q = get_relative_p(m, 1.1, 0.5)
    q0 = get_relative_p(2.0, 1.1, 0.5)
    ay = a.get_amp({"m": m})
    by = b.get_amp({"m": m}, {"|q|": q, "|q0|": q0})
    print(ay / by)
    check_frac(ay.numpy(), by.numpy())


if __name__ == "__main__":
    test_a()



import numpy as np

from .data import data_merge

has_uproot = True
try:
    import uproot3 as uproot
except ImportError as e:
    try:
        import uproot
    except ImportError as e:
        has_uproot = False
        print(e, "you should install `uproot` correctly for using this module")


def load_root_data(fnames):
    """load root file as dict"""
    if isinstance(fnames, str):
        fnames = [fnames]
    ret = {}
    root_file = [uproot.open(i) for i in fnames]
    keys = [i.allkeys() for i in root_file]
    common_keys = set()
    for i in keys:
        for j in i:
            pj = j.decode().split(";")[0]
            common_keys.add(pj)
    ret = {}
    for i in common_keys:
        data = []
        for j in root_file:
            data_i = load_Ttree(j.get(i))
            data.append(data_i)
        ret[i] = data_merge(*data)
    return ret


def load_Ttree(tree):
    """load TTree as dict"""
    ret = {}
    for i in tree.keys():
        arr = tree.get(i).array()
        if isinstance(arr, np.ndarray):
            ret[i.decode()] = arr
    return ret


def save_dict_to_root(dic, file_name, tree_name=None):
    """
    This function stores data arrays in the form of a dictionary into a root file.
    It provides a convenient interface to ``uproot``.

    :param dic: Dictionary of data
    :param file_name: String
    :param tree_name: String. By default it's "tree".
    """
    if file_name[-5:] == ".root":
        file_name = file_name[:-5]
    if isinstance(dic, dict):
        dic = [dic]
    if tree_name is None:
        tree_name = "DataTree"
    Ndic = len(dic)
    if isinstance(tree_name, list):
        assert len(tree_name) == Ndic
    else:
        t = []
        for i in range(Ndic):
            t.append(tree_name + str(i))
        tree_name = t

    with uproot.recreate(file_name + ".root") as f:
        for d, t in zip(dic, tree_name):
            branch_type = {}
            branch_data = {}
            for i in d:
                j = (
                    i.replace("(", "_")
                    .replace(")", "_")
                    .replace(" ", "_")
                    .replace("*", "star")
                    .replace("+", "p")
                    .replace("-", "m")
                )
                branch_data[j] = np.array(d[i])
                branch_type[j] = branch_data[j].dtype.name
            f[t] = uproot.newtree(branch_type)
            f[t].extend(branch_data)



import tensorflow as tf

from tf_pwa.amp.core import (
    HelicityDecay,
    Particle,
    get_relative_p2,
    register_decay,
    register_particle,
)
from tf_pwa.breit_wigner import Bprime_q2


@register_decay("LS-decay")
class ParticleDecayLS(HelicityDecay):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.same_phase = kwargs.get("same_phase", False)
        self.same_ratio = kwargs.get("same_ratio", False)

    def init_params(self):
        self.d = 3.0
        ls = self.get_ls_list()
        if len(ls) <= 0:
            print("no aviable ls", self, self.get_ls_list())
            return
        if self.same_ratio:
            if self.same_phase:
                self.g_ls = self.add_var(
                    "g_ls", is_complex=False, shape=(len(ls),)
                )
                for i in range(len(ls)):
                    self.g_ls.set_fix_idx(fix_idx=0, fix_vals=1.0)
            else:
                self.g_ls = self.add_var(
                    "g_ls", is_complex=True, shape=(len(ls),)
                )
                self.g_ls.set_fix_idx(fix_idx=0, fix_vals=(1.0, 0.0))
                self.g_ls.set_same_ratio()
        else:
            if self.same_phase:
                self.g_ls = self.add_var(
                    "g_ls", is_complex=False, shape=(len(ls),)
                )
                self.g_ls.set_fix_idx(fix_idx=0, fix_vals=1.0)
            else:
                self.g_ls = self.add_var(
                    "g_ls", is_complex=True, shape=(len(ls),)
                )
                self.g_ls.set_fix_idx(fix_idx=0, fix_vals=(1.0, 0.0))

    def get_barrier_factor2(self, mass, q2, q02, d):
        ls = self.get_ls_list()
        ls_amp = self.core.get_ls_amp(mass, ls, q2=q2, q02=q02, d=d)
        if self.ls_index is None:
            return tf.stack(ls_amp, axis=-1)
        return tf.stack([ls_amp[k] for k in self.ls_index], axis=-1)


class ParticleLS(Particle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        decay_params = kwargs.get("decay_params", {})
        self.decay_params = {"model": "LS-decay", **decay_params}

    def get_amp(self, *args, **kwargs):
        m = args[0]["m"]
        zeros = tf.zeros_like(m)
        ones = tf.ones_like(m)
        return tf.complex(ones, zeros)

    def get_ls_amp(self, m, ls, q2, q02, d=3):
        raise NotImplementedError


@register_particle("BWR_LS")
class ParticleBWRLS(ParticleLS):
    """

    Breit Wigner with split ls running width

    .. math::
        R_i (m) = \\frac{g_i}{m_0^2 - m^2 - im_0 \\Gamma_0 \\frac{\\rho}{\\rho_0} (\\sum_{i} g_i^2)}

    , :math:`\\rho = 2q/m`, the partial width factor is

    .. math::
        g_i = \\gamma_i \\frac{q^l}{q_0^l} B_{l_i}'(q,q_0,d)

    and keep normalize as

    .. math::
        \\sum_{i} \\gamma_i^2 = 1.

    The normalize is done by (:math:`\\cos \\theta_0, \\sin\\theta_0 \\cos \\theta_1, \\cdots, \\prod_i \\sin\\theta_i`)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.same_ratio = kwargs.get("same_ratio", True)
        self.same_phase = kwargs.get("same_phase", False)
        self.decay_params = {
            "same_ratio": self.same_ratio,
            "same_phase": self.same_phase,
            **self.decay_params,
        }
        self.theta = []

    def init_params(self):
        super().init_params()
        if getattr(self, "ls_list", None) is None:
            self.ls_list = self.decay[0].get_ls_list()
        self.theta = []
        for i in range(len(self.ls_list) - 1):
            self.theta.append(self.add_var(f"theta{i}"))

    def factor_gamma(self, ls):
        if len(ls) <= 1:
            return [1.0]
        f = 1.0
        ret = []
        for i in range(len(ls) - 1):
            theta_i = self.theta[i]()
            a = tf.cos(theta_i)
            ret.append(f * a)
            f = f * tf.sin(theta_i)
        ret.append(f)
        return ret

    def get_barrier_factor(self, ls, q2, q02, d):
        return [tf.sqrt(q2 / q02) ** i * Bprime_q2(i, q2, q02, d) for i in ls]

    def __call__(self, m):
        m0 = self.get_mass()
        m1 = self.decay[0].outs[0].get_mass()
        m2 = self.decay[0].outs[1].get_mass()
        ls = self.decay[0].get_ls_list()
        q2 = get_relative_p2(m, m1, m2)
        q02 = get_relative_p2(m0, m1, m2)
        return self.get_ls_amp(m, ls, q2, q02)

    def get_ls_amp(self, m, ls, q2, q02, d=3.0):
        assert all(i in self.ls_list for i in ls)
        ls = [i for i, j in self.ls_list]
        gammai = self.factor_gamma(ls)
        bf = self.get_barrier_factor(ls, q2, q02, d)
        total_gamma = [i * j for i, j in zip(gammai, bf)]
        m0 = self.get_mass()
        g0 = self.get_width()

        a = m0 * m0 - m * m
        b = (
            m0
            * g0
            * tf.sqrt(q2 / q02)
            * m
            / m0
            * sum([i * i for i in total_gamma])
        )
        dom = tf.complex(a, -b)

        ret = []
        for i in total_gamma:
            ret.append(tf.cast(i, dom.dtype) / dom)

        return ret



import operator
from functools import reduce

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.data import data_index, data_mask
from tf_pwa.histogram import Hist1D
from tf_pwa.model.model import sum_gradient


def binning_gradient(binning, amp, phsp, m_phsp, var):
    """
    calculate bin error for fit
    including error propagation (gradients) and weights
    """
    nfit = []
    for i in range(binning.shape[0] - 1):
        m_min = binning[i]
        m_max = binning[i + 1]
        phsp_i = data_mask(phsp, (m_phsp > m_min) & (m_phsp < m_max))
        weight_i = phsp_i.get("weight", 1.0)
        with tf.GradientTape() as tape:
            amp_i = amp(phsp_i)
            int_mc = tf.reduce_sum(weight_i * amp_i)
        grad = tape.gradient(int_mc, var, unconnected_gradients="zero")
        w_error2 = tf.reduce_sum(weight_i ** 2 * amp_i ** 2)
        grad = np.array([j.numpy() for j in grad])
        nfit.append((int_mc.numpy(), grad, w_error2.numpy()))
    count_i = np.array([i[0] for i in nfit])
    sum_int = np.sum(count_i)
    grads = np.array([i[1] for i in nfit])
    sum_int_grad = np.sum(grads, axis=0)
    w_error2 = np.array([i[2] for i in nfit]) / sum_int / sum_int
    grads = (grads - sum_int_grad * count_i[:, None] / sum_int) / sum_int
    return count_i / sum_int, grads, w_error2


def main():
    Nbins = 72
    config = ConfigLoader("config.yml")
    config.set_params("final_params.json")
    error_matrix = np.load("error_matrix.npy")

    idx = config.get_data_index("mass", "R_BC")

    datas = config.get_data("data")
    phsps = config.get_data("phsp")
    bgs = config.get_data("bg")
    amp = config.get_amplitude()
    var = amp.trainable_variables

    get_data = lambda x: data_index(x, idx).numpy()

    m_all = np.concatenate([get_data(i) for i in datas])

    m_min = np.min(m_all) - 0.1
    m_max = np.max(m_all) + 0.1
    binning = np.linspace(m_min, m_max, Nbins + 1)

    get_hist = lambda x, w: Hist1D.histogram(
        get_data(x), bins=Nbins, range=(m_min, m_max), weights=w
    )

    data_hist = [get_hist(i, i.get("weight")) for i in datas]
    bg_hist = [get_hist(i, np.abs(i.get("weight"))) for i in bgs]

    phsp_hist = []
    for dh, bh, phsp in zip(data_hist, bg_hist, phsps):
        m_phsp = data_index(phsp, idx).numpy()

        y_frac, grads, w_error2 = binning_gradient(
            binning, amp, phsp, m_phsp, var
        )
        error2 = np.einsum("ij,jk,ik->i", grads, error_matrix, grads)
        # error parameters and error from integration sample weights
        yerr = np.sqrt(error2 + w_error2)

        n_fit = dh.get_count() - bh.get_count()
        phsp_hist.append(n_fit * Hist1D(binning, y_frac, yerr))

    total_data = reduce(operator.add, data_hist)
    ax = plt.subplot2grid((4, 1), (0, 0), rowspan=3)
    total_data.draw(ax, label="data")
    total_data.draw_error(ax)
    total_bg = reduce(operator.add, bg_hist)
    total_bg.draw_bar(ax, label="back ground", color="grey", alpha=0.5)
    total_fit = reduce(operator.add, phsp_hist + bg_hist)
    total_fit.draw(ax, label="fit")
    total_fit.draw_error(ax)
    ax.set_ylim((0, None))
    ax.set_xlim((m_min, m_max))
    ax.legend()
    ax.set_ylabel(f"Events/ {(m_max-m_min)/Nbins:.3f} GeV")
    ax2 = plt.subplot2grid((4, 1), (3, 0), rowspan=1)
    (total_data - total_fit).draw_pull(ax2)
    plt.setp(ax.get_xticklabels(), visible=False)
    ax2.set_ylim((-5, 5))
    ax2.set_xlim((m_min, m_max))
    ax2.axhline(0, c="black", ls="-")
    ax2.axhline(-3, c="r", ls="--")
    ax2.axhline(3, c="r", ls="--")
    ax2.set_xlabel("$M(BC)$/GeV", loc="right")
    plt.savefig("fit_full_error.png")


if __name__ == "__main__":
    with tf.device("CPU"):
        main()



import warnings

from opt_einsum import contract, contract_path, get_symbol

from .tensorflow_wrapper import tf

# from pysnooper import snoop


class Einsum(object):
    def __init__(self, expr, shapes):
        self.expr = expr
        self.shapes = shapes

    def __call__(self, *args):
        return contract(self.expr, *args, backend="tensorflow")


def symbol_generate(base_map):
    if isinstance(base_map, dict):
        base_map = base_map.values()
    for i in range(100):
        symbol = get_symbol(i)
        if symbol not in base_map:
            yield symbol


def replace_ellipsis(expr, shapes):
    ret = expr
    idx = expr.split("->")[0].split(",")[0]
    extra = []
    if "..." in expr:
        extra_size = len(shapes[0]) - len(idx) + 3
        base_map = set(expr) - {".", "-", ">", ","}
        base_map = dict(enumerate(base_map))
        ig = symbol_generate(base_map)
        for i in range(extra_size):
            extra.append(next(ig))
        ret = expr.replace("...", "".join(extra))
    return ret, extra


def _get_order_bound_list(bd_dict, ord_dict, idx, left=0):
    if idx in ord_dict:
        return [ord_dict[idx]]
    assert idx in bd_dict, "not found"
    bd = bd_dict[idx]
    od = []
    for i in bd[left]:
        if i in ord_dict:
            od.append(ord_dict[i])
        else:
            od += _get_order_bound_list(bd_dict, ord_dict, i, left)
    return od


def ordered_indices(expr, shapes):
    """
    find a better order to reduce transpose.

    """
    ein_s = expr.split("->")
    final_index = ein_s[1]
    idx_input = ein_s[0].split(",")
    base_order = dict(zip(final_index, range(len(final_index))))
    max_i = len(expr)
    base_order["_min"] = -1
    base_order["_max"] = max_i

    combined_index = set("".join(idx_input)) - set(final_index)
    bound_dict = {}
    for i in combined_index:
        bound_dict[i] = ([], [])

    for i in combined_index:
        for j in idx_input:
            if i in j:
                pos = j.index(i)
                if pos - 1 >= 0:
                    bound_dict[i][0].append(j[pos - 1])
                else:
                    bound_dict[i][0].append("_min")
                if pos + 1 < len(j):
                    bound_dict[i][1].append(j[pos + 1])
                else:
                    bound_dict[i][1].append("_max")

    for i in bound_dict:
        left = max(_get_order_bound_list(bound_dict, base_order, i, 0))
        right = min(_get_order_bound_list(bound_dict, base_order, i, 1))
        if right > left:
            base_order[i] = left * 0.4 + right * 0.6
        else:
            base_order[i] = left + 0.01

    base_order = dict(sorted(base_order.items(), key=lambda x: x[0]))
    return base_order


def replace_none_in_shape(x, num=-1):
    # print(x.shape)
    shape = tuple([num if i is None else i for i in x])
    return shape


def remove_size1(expr, *args, extra=None):
    """remove order independent indices (size 1)"""
    if extra is None:
        extra = []
    sub = expr.split("->")[0].split(",")

    size_map = {}
    for idx, shape in zip(sub, args):
        for i, j in zip(idx, shape.shape):
            l = size_map.get(i, 1)
            if j is None or j >= l:
                size_map[i] = j

    remove_idx = []
    for i in size_map:
        if size_map[i] == 1 and i not in extra:
            remove_idx.append(i)

    idxs = expr.split("->")[0].split(",")
    idxs2 = []
    ret = []
    for idx, arg in zip(idxs, args):
        shape = []
        idx2 = []
        for i, j in zip(idx, arg.shape):
            if i not in remove_idx:
                shape.append(j)
                idx2.append(i)
        ret.append(tf.reshape(arg, replace_none_in_shape(shape, -1)))
        idxs2.append("".join(idx2))

    final_idx = expr.split("->")[1]
    for i in remove_idx:
        final_idx = final_idx.replace(i, "")
    expr2 = ",".join(idxs2) + "->" + final_idx

    return expr2, ret, size_map


def einsum(expr, *args, **kwargs):
    shapes = [replace_none_in_shape(i.shape, 10000) for i in args]
    expr, extra = replace_ellipsis(expr, shapes)
    path, path_info = contract_path(
        expr, *shapes, shapes=True, optimize="auto"
    )
    final_idx = expr.split("->")[1]
    expr2, args, size_map = remove_size1(expr, *args, extra=extra)
    final_shape = [size_map[i] for i in final_idx]
    base_order = ordered_indices(expr2, shapes)
    ein_s = expr2.split("->")
    final_index = ein_s[1]
    idxs = ein_s[0].split(",")

    data = list(args)
    in_idx = list(idxs)
    for idx in path:
        part_data = [data[i] for i in idx]
        part_in_idx = [in_idx[i] for i in idx]
        for i in sorted(idx)[::-1]:
            del data[i]
            del in_idx[i]
        out_idx = set("".join(part_in_idx)) & set(
            final_index + "".join(in_idx)
        )
        out_idx = "".join(sorted(out_idx, key=lambda x: base_order[x]))
        in_idx.append(out_idx)
        expr_i = "{}->{}".format(",".join(part_in_idx), out_idx)
        result = tensor_einsum_reduce_sum(expr_i, *part_data, order=base_order)
        data.append(result)
    return tf.reshape(data[0], replace_none_in_shape(final_shape, -1))


def tensor_einsum_reduce_sum(expr, *args, order):
    """
    "abe,bcf->acef"  =reshape=> "ab1e1,1bc1f->acef" =product=> "abcef->acef" =reduce_sum=> "acef"
    """
    ein_s = expr.split("->")
    final_index = ein_s[1]
    idxs = ein_s[0].split(",")
    for i in idxs:
        if len(set(i)) != len(i):  # inner product
            warnings.warn("inner product")
            return tf.einsum(expr, *args)

    require_order = sorted(set(ein_s[0]) - {","}, key=lambda x: order[x])

    # transpose
    t_args = []
    n_args = len(idxs)

    def args_it(it):
        i, j = idxs[it], args[it]
        sorted_idx = sorted(i, key=lambda x: order[x])
        if list(i) == sorted_idx:
            return j
        else:
            trans = [i.index(k) for k in sorted_idx]
            return tf.transpose(j, trans)

    t_args = [args_it(it) for it in range(n_args)]
    # reshape
    sum_idx = set(require_order) - set(final_index)
    sum_idx_idx = [i for i, j in enumerate(require_order) if j in sum_idx]
    shapes = [i.shape for i in args]

    def expand_shape_it(idx, shape):
        shape_dict = dict(zip(idx, shape))
        ex_shape = [shape_dict.get(i, 1) for i in require_order]
        return ex_shape

    expand_shapes = [
        expand_shape_it(idx, shape) for idx, shape in zip(idxs, shapes)
    ]

    # print(expr, order, expand_shapes)
    s_args = [
        tf.reshape(j, replace_none_in_shape(i, -1))
        for i, j in zip(expand_shapes, t_args)
    ]

    # product
    ret_1 = s_args.pop()
    while len(s_args) > 0:
        ret_1 = ret_1 * s_args.pop()

    # reduce_sum
    ret = tf.reduce_sum(ret_1, axis=sum_idx_idx)
    return ret



import operator
from functools import reduce

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf

from tf_pwa.config_loader import ConfigLoader, MultiConfig
from tf_pwa.data import data_index, data_mask
from tf_pwa.histogram import Hist1D, WeightedData


def sum_hist(x):
    return reduce(operator.add, x)


def plot_mass(
    amp, datas, bgs, phsps, get_data, name, Nbins, xrange=None, unit="GeV"
):

    # determine histogram range
    if xrange is None:
        m_all = np.concatenate([get_data(i) for i in datas])
        m_min, m_max = np.min(m_all), np.max(m_all)
        delta_m = (m_max - m_min) / Nbins
        m_min, m_max = m_min - 3 * delta_m, m_max + 3 * delta_m
    else:
        m_min, m_max = xrange

    binning = np.linspace(m_min, m_max, Nbins + 1)

    get_hist = lambda x, w: Hist1D.histogram(
        get_data(x), bins=Nbins, range=(m_min, m_max), weights=w
    )
    # using more bins to smooth
    get_hist2 = lambda x, w: WeightedData(
        get_data(x), bins=Nbins * 2, range=(m_min, m_max), weights=w * 2
    )
    # get data and background histogram
    data_hist = [get_hist(i, i.get("weight")) for i in datas]
    bg_hist = [get_hist(i, np.abs(i.get("weight"))) for i in bgs]
    phsp_hist = []
    pw_hist = []
    for dh, bh, phsp in zip(data_hist, bg_hist, phsps):
        # using amplitude square as weights for fit
        amp_i = phsp.get("weight", 1.0) * amp(phsp)
        int_mc = tf.reduce_sum(amp_i)
        # normalize to signal number
        scale = (dh.get_count() - bh.get_count()) / int_mc
        y_frac = amp_i * scale
        phsp_hist.append(get_hist(phsp, y_frac.numpy()))

        # calculate partial wave weights
        pw = amp.partial_weight(phsp)
        tmp = []
        for i in pw:
            y_frac = phsp.get("weight", 1.0) * i * scale
            tmp.append(get_hist2(phsp, y_frac.numpy()))
        pw_hist.append(tmp)

    total_data = sum_hist(data_hist)
    total_bg = sum_hist(bg_hist)
    total_fit = sum_hist(phsp_hist + bg_hist)

    # plot data and fit histogram
    ax = plt.subplot2grid((4, 1), (0, 0), rowspan=3)
    total_data.draw(ax, label="data")
    total_data.draw_error(ax)
    total_bg.draw_bar(ax, label="back ground", color="grey", alpha=0.5)
    total_fit.draw(ax, label="fit")
    # total_fit.draw_error(ax)
    for hi, dec in zip(zip(*pw_hist), amp.decay_group):
        h = sum_hist(hi)
        res = dec[0].outs
        name_i = " ".join([getattr(i, "display", str(i)) for i in res])
        # h.draw(ax, label=name_i)
        h.draw_kde(ax, label=name_i)
    # plot legend
    # ax.legend()
    # plot pull
    ax2 = plt.subplot2grid((4, 1), (3, 0), rowspan=1)
    (total_data - total_fit).draw_pull(ax2)
    ax2.axhline(0, c="black", ls="-")
    ax2.axhline(-3, c="r", ls="--")
    ax2.axhline(3, c="r", ls="--")
    ax.grid()
    ax2.grid()
    # set some axis attributes
    ax.set_ylim((0, None))
    ax.set_xlim((m_min, m_max))
    ax.set_ylabel(f"Events/ {(m_max-m_min)/Nbins:.3f} {unit}")
    plt.setp(ax.get_xticklabels(), visible=False)
    ax2.set_ylim((-5, 5))
    ax2.set_xlim((m_min, m_max))
    ax2.set_xlabel(f"${name}$/{unit}", loc="right")
    # save figure
    plt.savefig(f"{name}.png", dpi=300)


def main():
    Nbins = 64
    config = ConfigLoader("config.yml")
    # config = MultiConfig(["config.yml"]).configs[0]
    config.set_params("final_params.json")
    name = "R_BC"
    idx = config.get_data_index("mass", "R_BC")
    # idx_costheta = (*config.get_data_index("angle", "DstD/D*"), "beta")

    datas, phsps, bgs, _ = config.get_all_data()
    amp = config.get_amplitude()
    get_data = lambda x: data_index(x, idx).numpy()
    # get_data = lambda x: np.cos(data_index(x, idx_costheta).numpy())
    plot_mass(amp, datas, bgs, phsps, get_data, name, Nbins)


if __name__ == "__main__":
    main()



"""
Examples for config.yml file
----------------------------

Configuration file `config.yml` use YAML (https://yaml.org) format to describe decay process.

The main parts of config.yml is `decay` and `particle`.

The `decay` part describe the particle (or an id of a list of particle)
decay into which particles, it can be a list of a list of list.
A list means that there is ony one decay mode, A list of list is the list of possible decay mode.
The list item can be the particle name (or a dict to describe the decay parameters).
All name should appear in `particle` part.

The `particle` part describe the parameters of particles.
There are two special parts `$top` and `$finals` describe the top and finals particles.
The other parts are lists of particle name or dicts of particle parameters.
The list is the same type particle in decay part.
The dict is the parameters of the particle name.
"""

config_str = """

decay:
    A:
       - [R1, B]
       - [R2, C]
       - [R3, D]
    R1: [C, D]
    R2: [B, D]
    R3: [B, C]

particle:
    $top:
       A: { mass: 1.86, J: 0, P: -1}
    $finals:
       B: { mass: 0.494, J: 0, P: -1}
       C: { mass: 0.139, J: 0, P: -1}
       D: { mass: 0.139, J: 0, P: -1}
    R1: [ R1_a, R1_b ]
    R1_a: { mass: 0.7, width: 0.05, J: 1, P: -1}
    R1_b: { mass: 0.5, width: 0.05, J: 0, P: +1}
    R2: { mass: 0.824, width: 0.05, J: 0, P: +1}
    R3: { mass: 0.824, width: 0.05, J: 0, P: +1}

"""

# %%
# The config file can be loaded by `yaml` library.
#

import matplotlib.pyplot as plt
import yaml

from tf_pwa.config_loader import ConfigLoader
from tf_pwa.histogram import Hist1D

config = ConfigLoader(yaml.full_load(config_str))

# %%
# We set parameters to a blance value. And we can generate some toy data and calclute the weights
#

input_params = {
    "A->R1_a.BR1_a->C.D_total_0r": 6.0,
    "A->R1_b.BR1_b->C.D_total_0r": 1.0,
    "A->R2.CR2->B.D_total_0r": 2.0,
    "A->R3.DR3->B.C_total_0r": 1.0,
}
config.set_params(input_params)

data = config.generate_toy(1000)
phsp = config.generate_phsp(10000)

# You can also fit the data fit to the data
fit_result = config.fit([data], [phsp])
err = config.get_params_error(fit_result, [data], [phsp])

# %%
# we can see that thre fit results consistant with inputs, the first one is fixed.

for var in input_params:
    print(
        f"in: {input_params[var]} => out: {fit_result.params[var]} +/- {err.get(var, 0.)}"
    )

# %%
# We can use the amplitude to plot the fit results

amp = config.get_amplitude()
weight = amp(phsp)
partial_weight = amp.partial_weight(phsp)

# %%
# We can plot the data, Hist1D include some plot method base on matplotlib.

data_hist = Hist1D.histogram(
    data.get_mass("(C, D)"), bins=60, range=(0.25, 1.45)
)

mass_phsp = phsp.get_mass("(C, D)")
phsp_hist = Hist1D.histogram(
    mass_phsp, weights=weight, bins=60, range=(0.25, 1.45)
)
scale = phsp_hist.scale_to(data_hist)

pw_hist = []
for w in partial_weight:
    # here we used more bins for a smooth plot
    hist = Hist1D.histogram(
        mass_phsp, weights=w, bins=60 * 2, range=(0.25, 1.45)
    )
    pw_hist.append(scale * hist * 2)

# %%
# Then we can plot the histogram into matplotlib

for hist, dec in zip(pw_hist, config.get_decay()):
    hist.draw_kde(label=str(dec))
phsp_hist.draw(label="total amplitude")
data_hist.draw_error(label="toy data", color="black")

plt.legend()
plt.ylim((0, None))
plt.xlabel("M(CD)/ GeV")
plt.ylabel("Events/ 20 MeV")
plt.show()



from itertools import product

import matplotlib.pyplot as plt
import numpy as np
import sympy as sym
from sympy import Symbol, simplify, trigsimp
from sympy.physics.quantum.cg import CG
from sympy.physics.quantum.spin import WignerD

from tf_pwa.config_loader import ConfigLoader


def spin_int(i):
    """use it for half integer in spins"""
    return sym.S(int(abs(i) * 2 + 0.001)) / 2 * sym.sign(i)


def get_decay_part(decay, ls, lambda_list, symbol_list):
    """
    .. math::
        \\sqrt{\\frac{ 2 l + 1 }{ 2 j_a + 1 }}
        \\langle j_b, j_c, \\lambda_b, - \\lambda_c | s, \\lambda_b - \\lambda_c \\rangle
        \\langle l, s, 0, \\lambda_b - \\lambda_c | j_a, \\lambda_b - \\lambda_c \\rangle

    .. math::
        D_{\\lambda_a, \\lambda_b - \\lambda_c}^{J_{A}*} (\\phi, \\theta, 0)

    """
    a = decay.core
    b, c = decay.outs
    l, s = ls
    lambda_list = {k: spin_int(v) for k, v in lambda_list.items()}
    delta = lambda_list[b] - lambda_list[c]
    if abs(delta) > a.J:
        return 0
    d_part = WignerD(
        spin_int(a.J),
        lambda_list[a],
        delta,
        symbol_list[decay]["alpha"],
        symbol_list[decay]["beta"],
        0,
    )
    # print(lambda_list)
    cg_part = sym.sqrt((2 * l + 1)) / sym.sqrt((2 * spin_int(a.J) + 1))
    cg_part = cg_part * CG(l, 0, spin_int(s), delta, spin_int(a.J), delta)
    cg_part = cg_part * CG(
        spin_int(b.J),
        lambda_list[b],
        spin_int(c.J),
        -lambda_list[c],
        spin_int(s),
        delta,
    )
    return simplify(d_part.conjugate() * cg_part)


def get_angle_distrubution_single(decay_chain, ls_list):
    ret = {}
    symbol_list = {
        decay: {
            "alpha": Symbol(f"phi{i}", real=True),
            "beta": Symbol(f"theta{i}", real=True),
        }
        for i, decay in enumerate(decay_chain)
    }
    out_particle = [decay_chain.top] + decay_chain.outs
    out_helicity = [i.spins for i in out_particle]
    inner_helicity = [i.spins for i in decay_chain.inner]
    for i in product(*out_helicity):
        lambda_list_out = dict(zip(out_particle, i))
        ret_part = 0
        for j in product(*inner_helicity):
            lambda_list = dict(zip(decay_chain.inner, j))
            lambda_list.update(lambda_list_out)
            tmp = 1
            for decay, ls in zip(decay_chain, ls_list):
                tmp = tmp * get_decay_part(decay, ls, lambda_list, symbol_list)
            ret_part = ret_part + tmp
        ret[i] = ret_part.expand(complex=True)
    return ret


def get_angle_distrubution(decay_chain):
    ls_com = []
    for i in decay_chain:
        ls_com.append(i.get_ls_list())

    ret = {}
    for i in product(*ls_com):
        ret[i] = get_angle_distrubution_single(decay_chain, i)
    return ret


def get_projection(f_theta, var):
    var = Symbol(var, real=True)
    args = f_theta.free_symbols
    inte_args = [i for i in args if i != var]
    inte_params = []
    for i in inte_args:
        if str(i).startswith("theta"):
            inte_params.append((i, 0, sym.pi))
            # d cos(theta) = sin(theta) d theta
            f_theta = f_theta * sym.sin(i)
        else:
            inte_params.append((i, -sym.pi, sym.pi))
    # projection in theta
    if len(inte_args) > 0:
        f_theta1 = sym.integrate(f_theta, *inte_params)
    else:
        f_theta1 = f_theta

    if str(var).startswith("theta"):
        normal = sym.integrate(f_theta1 * sym.sin(var), (var, 0, sym.pi))
    else:
        var_range = (-sym.pi, sym.pi)
        normal = sym.integrate(f_theta1, (var, -sym.pi, sym.pi))

    f_theta1 = simplify(f_theta1 / normal)
    return f_theta1


def plot_theta(f_name, f_theta, var):
    """ plot phi and cos theta """
    f_theta1 = get_projection(f_theta, var)
    var = Symbol(var, real=True)

    if str(var).startswith("theta"):
        var_range = (0, np.pi)
        trans = np.cos
    else:
        var_range = (-np.pi, np.pi)
        trans = lambda x: x

    print(f_theta1)

    f = sym.lambdify((var,), f_theta1.evalf(), "numpy")
    theta = np.linspace(var_range[0], var_range[1], 1000)
    x = trans(theta)
    y = f(theta) + np.zeros_like(x)

    plt.clf()
    plt.title("${}$".format(sym.latex(f_theta1)))
    plt.plot(x, y)
    plt.xlim((np.min(x), np.max(x)))
    plt.ylim((0, None))
    plt.savefig(f_name)


def main():
    sym.init_printing()
    config = ConfigLoader("config.yml")
    decay_group = config.get_decay()
    decay_chain = list(decay_group)[0]
    print(decay_chain)
    angle = get_angle_distrubution(decay_chain)
    ret = []
    for k, v in angle.items():
        for k2, v2 in v.items():
            f_theta = simplify(v2 * v2.conjugate())
            print("ls: ", k)
            print("  lambda:", k2)
            print("    :", f_theta)
            ret.append(f_theta)
            # break

    f_theta = ret[0]
    plot_theta("costheta.png", f_theta, "theta1")
    plot_theta("costheta2.png", f_theta, "theta2")
    plot_theta("phi2.png", f_theta, "phi2")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import os.path
import sys

import matplotlib.colors as mcolors
import matplotlib.patches as mpathes
import matplotlib.pyplot as plt
import numpy as np

from tf_pwa.adaptive_bins import AdaptiveBound, cal_chi2
from tf_pwa.angle import kine_max, kine_min
from tf_pwa.config_loader import ConfigLoader
from tf_pwa.data import data_index, data_to_numpy

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


def cal_chi2_config(
    config, adapter, data, phsp, data_idx, bg=None, data_cut=None
):
    if data_cut is None:
        data_cut = np.array([data_index(data, idx) for idx in data_idx])
    amp_weight = config.get_amplitude()(phsp).numpy()
    phsp_cut = np.array([data_index(phsp, idx) for idx in data_idx])
    phsp_slice = np.concatenate(
        [np.array(phsp_cut ** 2), [amp_weight]], axis=0
    )
    phsps = adapter.split_data(phsp_slice)
    datas = adapter.split_data(data_cut ** 2)
    bound = adapter.get_bounds()
    if bg is not None:
        bg_weight = config._get_bg_weight(display=False)[0][0]
        bg_cut = np.array([data_index(bg, idx) for idx in data_idx])
        bgs = adapter.split_data(bg_cut ** 2)
        int_norm = (
            data_cut.shape[-1] - bg_cut.shape[-1] * bg_weight
        ) / np.sum(amp_weight)
    else:
        int_norm = data_cut.shape[-1] / np.sum(amp_weight)
    print("int norm:", int_norm)
    numbers = []
    for i, bnd in enumerate(bound):
        min_x, min_y = bnd[0]
        max_x, max_y = bnd[1]
        ndata = datas[i].shape[-1]
        nmc = np.sum(phsps[i][2]) * int_norm
        if bg is not None:
            nmc += bgs[i].shape[-1] * bg_weight
        numbers.append((ndata, nmc))
    return cal_chi2(numbers, config.get_ndf())


def draw_dalitz(data_cut, bound):
    fig, ax = plt.subplots()
    my_cmap = plt.get_cmap("jet")
    # my_cmap.set_under('w', 1)

    for i, bnd in enumerate(bound):
        min_x, min_y = bnd[0]
        max_x, max_y = bnd[1]
        rect = mpathes.Rectangle(
            (min_x, min_y),
            max_x - min_x,
            max_y - min_y,
            linewidth=1,
            facecolor="none",
            edgecolor="black",
        )  # cmap(weights[i]/max_weight))
        ax.add_patch(rect)

    ah = ax.hist2d(
        data_cut[0] ** 2, data_cut[1] ** 2, bins=50, norm=mcolors.LogNorm()
    )
    # ax.scatter(data_cut[0]**2, data_cut[1]**2, s=1, c="red")
    ## using your own mass
    m0, mb, mc, md = 5.27926, 2.01026, 1.86961, 0.49368
    # print(ah)
    sbc_min, sbc_max = (mb + mc) ** 2, (m0 - md) ** 2
    sbd_min, sbd_max = (mb + md) ** 2, (m0 - mc) ** 2
    sbc = np.linspace(sbc_min, sbc_max, 1000)
    ax.plot(sbc, kine_max(sbc, m0, mc, mb, md), color="grey")
    ax.plot(sbc, kine_min(sbc, m0, mc, mb, md), color="grey")

    ax.set_xlim((sbc_min, sbc_max))
    ax.set_ylim((sbd_min, sbd_max))
    ax.set_xlabel("$M_{BC}$")
    ax.set_ylabel("$M_{BD}$")
    fig.colorbar(ah[-1])
    plt.savefig("figure/mbc_mbd_adaptive.png", dpi=200)


def main():
    config = ConfigLoader("config.yml")
    data = config.get_data("data")[0]
    phsp = config.get_data("phsp")[0]
    bg = config.get_data("bg")[0]
    config.set_params("final_params.json")

    mbc_idx = config.get_data_index(
        "mass", "R_BC"
    )  # ("particle", "(D, K)", "m")
    mbd_idx = config.get_data_index(
        "mass", "R_BD"
    )  # ("particle", "(D0, K, pi)", "m")
    data_idx = [mbc_idx, mbd_idx]

    data_cut = np.array([data_index(data, idx) for idx in data_idx])
    adapter = AdaptiveBound(data_cut ** 2, [[2, 2], [3, 3], [2, 2]])
    bound = adapter.get_bounds()

    cal_chi2_config(
        config, adapter, data, phsp, data_idx, bg=bg, data_cut=data_cut
    )
    draw_dalitz(data_cut, bound)


if __name__ == "__main__":
    main()



import copy

import yaml


def load_config(file_name, share_dict=None):
    if share_dict is None:
        share_dict = {}
    if isinstance(file_name, dict):
        return copy.deepcopy(file_name)
    if isinstance(file_name, str):
        if file_name in share_dict:
            return load_config(share_dict[file_name])
        with open(file_name) as f:
            ret = yaml.load(f, yaml.FullLoader)
        return ret
    raise TypeError("not support config {}".format(type(file_name)))


class BaseConfig(object):
    def __init__(self, file_name, share_dict=None):
        self.config = load_config(file_name, share_dict)

    def __getitem__(self, key):
        return self.config.get(key)

    def __setitem__(self, key, value):
        self.config[key] = value



import json
import time

import numpy as np
import tensorflow as tf
from scipy.optimize import BFGS, basinhopping, minimize

from .fit_improve import Cached_FG
from .fit_improve import minimize as my_minimize
from .utils import time_print


class LargeNumberError(ValueError):
    pass


def fit_minuit(fcn, bounds_dict={}, hesse=True, minos=False, **kwargs):
    try:
        import iminuit
    except ImportError:
        raise RuntimeError(
            "You haven't installed iminuit so you can't use Minuit to fit."
        )
    """

    :param fcn:
    :param bounds_dict:
    :param hesse:
    :param minos:
    :return:
    """
    if int(iminuit.__version__[0]) < 2:
        return fit_minuit_v1(
            fcn, bounds_dict=bounds_dict, hesse=hesse, minos=minos, **kwargs
        )
    return fit_minuit_v2(
        fcn, bounds_dict=bounds_dict, hesse=hesse, minos=minos, **kwargs
    )


def fit_minuit_v1(fcn, bounds_dict={}, hesse=True, minos=False, **kwargs):

    """

    :param fcn:
    :param bounds_dict:
    :param hesse:
    :param minos:
    :return:
    """
    from iminuit import Minuit

    var_args = {}
    var_names = fcn.vm.trainable_vars
    for i in var_names:
        var_args[i] = fcn.vm.get(i)
        if i in bounds_dict:
            var_args["limit_{}".format(i)] = bounds_dict[i]
        var_args["error_" + i] = 0.1

    f_g = Cached_FG(fcn.nll_grad)
    m = Minuit(
        f_g.fun,
        name=var_names,
        errordef=0.5,
        grad=f_g.grad,
        print_level=2,
        use_array_call=True,
        **var_args,
    )
    print("########## begin MIGRAD")
    now = time.time()
    m.migrad()  # (ncall=10000))#,precision=5e-7))
    print("MIGRAD Time", time.time() - now)
    if hesse:
        print("########## begin HESSE")
        now = time.time()
        m.hesse()
        print("HESSE Time", time.time() - now)
    if minos:
        print("########## begin MINOS")
        now = time.time()
        m.minos()  # (var="")
        print("MINOS Time", time.time() - now)
    ndf = len(m.list_of_vary_param())
    ret = FitResult(
        dict(m.values), fcn, m.fval, ndf=ndf, success=m.migrad_ok()
    )
    ret.set_error(dict(m.errors))
    return ret


def fit_minuit_v2(fcn, bounds_dict={}, hesse=True, minos=False, **kwargs):

    """

    :param fcn:
    :param bounds_dict:
    :param hesse:
    :param minos:
    :return:
    """
    from iminuit import Minuit

    var_args = {}
    var_names = fcn.vm.trainable_vars
    x0 = []
    for i in var_names:
        x0.append(fcn.vm.get(i))
        var_args[i] = fcn.vm.get(i)
        if i in bounds_dict:
            var_args["limit_{}".format(i)] = bounds_dict[i]
        # var_args["error_" + i] = 0.1

    f_g = Cached_FG(fcn.nll_grad)
    m = Minuit(
        f_g.fun,
        np.array(x0),
        name=var_names,
        grad=f_g.grad,
    )
    m.strategy = 0
    for i in var_names:
        if i in bounds_dict:
            m.limits[i] = bounds_dict[i]
    m.errordef = 0.5
    m.print_level = 2
    print("########## begin MIGRAD")
    now = time.time()
    m.migrad()  # (ncall=10000))#,precision=5e-7))
    print("MIGRAD Time", time.time() - now)
    if hesse:
        print("########## begin HESSE")
        now = time.time()
        m.hesse()
        print("HESSE Time", time.time() - now)
    if minos:
        print("########## begin MINOS")
        now = time.time()
        m.minos()  # (var="")
        print("MINOS Time", time.time() - now)
    ndf = len(var_names)
    ret = FitResult(
        dict(zip(var_names, m.values)), fcn, m.fval, ndf=ndf, success=m.valid
    )
    # print(m.errors)
    ret.set_error(dict(zip(var_names, m.errors)))
    return ret


def fit_root_fitter(fcn):
    from array import array

    import ROOT

    var_names = fcn.vm.trainable_vars
    x0 = []
    for i in var_names:
        x0.append(fcn.vm.get(i))
    f_g = Cached_FG(fcn.nll_grad)

    class MyMultiGenFCN(ROOT.Math.IMultiGenFunction):
        def NDim(self):
            return len(x0)

        def DoEval(self, x):
            return f_g.fun(x)

        def Clone(self):
            x = MyMultiGenFCN()
            ROOT.SetOwnership(x, False)
            return x

    fitter = ROOT.Fit.Fitter()
    myMultiGenFCN = MyMultiGenFCN()
    params = array("d", x0)
    fitter.FitFCN(myMultiGenFCN, params)
    fit_result = fitter.Result()
    x = dict(zip(var_names, [fit_result.Parameter(i) for i in range(len(x0))]))
    xerr = dict(zip(var_names, [fit_result.Error(i) for i in range(len(x0))]))
    ndf = len(var_names)
    ret = FitResult(
        x, fcn, fit_result.MinFcnValue(), ndf=ndf, success=fit_result.IsValid()
    )
    ret.set_error(xerr)
    return ret


def fit_scipy(
    fcn,
    method="BFGS",
    bounds_dict={},
    check_grad=False,
    improve=False,
    maxiter=None,
):
    """

    :param fcn:
    :param method:
    :param bounds_dict:
    :param kwargs:
    :return:
    """
    args_name = fcn.vm.trainable_vars
    x0 = []
    bnds = []
    for name, i in zip(args_name, fcn.vm.trainable_variables):
        x0.append(i.numpy())
        if name in bounds_dict:
            bnds.append(bounds_dict[name])
        else:
            bnds.append((None, None))

    points = []
    nlls = []
    hess_inv = None
    now = time.time()
    if maxiter is None:
        maxiter = max(100 * len(x0), 2000)
    min_nll = 0.0
    ndf = 0
    # maxiter = 0
    def v_g2(x0):
        f_g = fcn.vm.trans_fcn_grad(fcn.nll_grad)
        nll, gs0 = f_g(x0)
        gs = []
        for i, name in enumerate(args_name):
            x0[i] += 1e-5
            nll0, _ = f_g(x0)
            x0[i] -= 2e-5
            nll1, _ = f_g(x0)
            x0[i] += 1e-5
            gs.append((nll0 - nll1) / 2e-5)
        return nll, np.array(gs)

    if check_grad:
        print("checking gradients ...")
        f_g = fcn.vm.trans_fcn_grad(fcn.nll_grad)
        nll, gs0 = f_g(x0)
        _, gs = v_g2(x0)
        for i, name in enumerate(args_name):
            print(args_name[i], gs[i], gs0[i])

    if method in ["BFGS", "CG", "Nelder-Mead", "test"]:

        def callback(x):
            if np.fabs(x).sum() > 1e7:
                x_p = dict(zip(args_name, x))
                raise LargeNumberError("x too large: {}".format(x_p))
            points.append(fcn.vm.get_all_val())
            nlls.append(float(fcn.cached_nll))
            # if len(nlls) > maxiter:
            #    with open("fit_curve.json", "w") as f:
            #        json.dump({"points": points, "nlls": nlls}, f, indent=2)
            #    pass  # raise Exception("Reached the largest iterations: {}".format(maxiter))
            print(fcn.cached_nll)

        # bd = Bounds(bnds)
        fcn.vm.set_bound(bounds_dict)

        f_g = fcn.vm.trans_fcn_grad(fcn.nll_grad)
        x0 = np.array(fcn.vm.get_all_val(True))
        # print(x0, fcn.vm.get_all_dic())
        # s = minimize(f_g, x0, method='trust-constr', jac=True, hess=BFGS(), options={'gtol': 1e-4, 'disp': True})
        if method == "test":
            try:
                s = my_minimize(
                    f_g,
                    x0,
                    method=method,
                    jac=True,
                    callback=callback,
                    options={"disp": 1, "gtol": 1e-3, "maxiter": maxiter},
                )
            except LargeNumberError:
                return except_result(fcn, x0.shape[0])
        else:
            try:
                s = minimize(
                    f_g,
                    x0,
                    method=method,
                    jac=True,
                    callback=callback,
                    options={"disp": 1, "gtol": 1e-3, "maxiter": maxiter},
                )
            except LargeNumberError:
                return except_result(fcn, x0.shape[0])

        while improve and not s.success:
            min_nll = s.fun
            maxiter -= s.nit
            s = minimize(
                f_g,
                s.x,
                method=method,
                jac=True,
                callback=callback,
                options={"disp": 1, "gtol": 1e-2, "maxiter": maxiter},
            )
            if hasattr(s, "hess_inv"):
                edm = np.dot(np.dot(s.hess_inv, s.jac), s.jac)
            else:
                break
            if edm < 1e-5:
                print("edm: ", edm)
                s.message = "Edm allowed"
                break
            if abs(s.fun - min_nll) < 1e-3:
                break
        fcn.vm.set_trans_var(s.x)  # make sure fit results same as variable
        print(s)
        # xn = s.x  # fcn.vm.get_all_val()  # bd.get_y(s.x)

        # fcn.vm.set_all(s.x, True)
        ndf = s.x.shape[0]
        min_nll = s.fun
        success = s.success
        hess_inv = fcn.vm.trans_error_matrix(s.hess_inv, s.x)
        fcn.vm.remove_bound()

        xn = fcn.vm.get_all_val()
    elif method in ["L-BFGS-B"]:

        def callback(x):
            if np.fabs(x).sum() > 1e7:
                x_p = dict(zip(args_name, x))
                raise LargeNumberError("x too large: {}".format(x_p))
            points.append([float(i) for i in x])
            nlls.append(float(fcn.cached_nll))

        try:

            s = minimize(
                fcn.nll_grad,
                np.array(x0),
                method=method,
                jac=True,
                bounds=bnds,
                callback=callback,
                options={
                    "disp": 1,
                    "maxcor": 50,
                    "ftol": 1e-15,
                    "maxiter": maxiter,
                },
            )
        except LargeNumberError:
            return except_result(fcn, len(x0))
        xn = s.x
        fcn.vm.set_var(xn)
        print(s)
        ndf = s.x.shape[0]
        min_nll = s.fun
        success = s.success
    elif method in ["Newton-CG", "trust-krylov", "trust-ncg", "trust-exact"]:
        fcn.vm.set_bound(bounds_dict)
        return fit_newton_cg(fcn, method, False)
    elif method in ["Newton-CG-p", "trust-krylov-p", "trust-ncg-p"]:
        fcn.vm.set_bound(bounds_dict)
        return fit_newton_cg(fcn, method[:-2], True)
    elif method in ["iminuit"]:
        m = fit_minuit(fcn)
        return m
    elif method in ["root"]:
        m = fit_root_fitter(fcn)
        return m
    else:
        raise Exception("unknown method")
    if check_grad:
        print("checking gradients ...")
        f_g = fcn.vm.trans_fcn_grad(fcn.nll_grad)
        _, gs0 = f_g(xn)
        gs = []
        for i, name in enumerate(args_name):
            xn[i] += 1e-5
            nll0, _ = f_g(xn)
            xn[i] -= 2e-5
            nll1, _ = f_g(xn)
            xn[i] += 1e-5
            gs.append((nll0 - nll1) / 2e-5)
            print(args_name[i], gs[i], gs0[i])
    params = fcn.get_params()  # vm.get_all_dic()
    return FitResult(
        params, fcn, min_nll, ndf=ndf, success=success, hess_inv=hess_inv
    )


def fit_newton_cg(
    fcn, method="Newton-CG", use_hessp=False, check_hess=False, gtol=1e-4
):
    vm = fcn.vm

    points = []

    def callback(x):
        points.append(fcn.vm.get_all_val())
        # if len(nlls) > maxiter:
        #    with open("fit_curve.json", "w") as f:
        #        json.dump({"points": points, "nlls": nlls}, f, indent=2)
        #    pass  # raise Exception("Reached the largest iterations: {}".format(maxiter))
        print(fcn.cached_nll)

    # bd = Bounds(bnds)

    f_g = vm.trans_fcn_grad(fcn.nll_grad)
    if use_hessp:
        hessp = vm.trans_grad_hessp(fcn.grad_hessp)
    else:
        hess = vm.trans_f_grad_hess(fcn.nll_grad_hessian)
        hess = time_print(hess)
    f_g = Cached_FG(f_g)

    x0 = np.array(vm.get_all_val(True))

    if check_hess:
        # check if hessp works well
        # hess(x0, x0)
        gs = []
        for i, _ in enumerate(x0):
            x0[i] += 1e-3
            _, g1 = f_g(x0)
            x0[i] -= 2e-3
            _, g2 = f_g(x0)
            x0[i] += 1e-3
            gs.append((g1 - g2) / 2e-3)
        gs = np.array(gs)

        # print(gs)
        if use_hessp:
            p = np.random.random(x0.shape)
            print("check", hessp(x0, p)[1], "==", np.dot(gs, p))
            p = np.random.random(x0.shape)
            print("check", hessp(x0, p)[1], "==", np.dot(gs, p))
        else:
            print(hess(x0)[2] - gs)

    if use_hessp:
        s = minimize(
            f_g, x0, jac=True, hessp=lambda x, p: hessp(x, p)[1], method=method
        )
    else:
        s = minimize(
            f_g, x0, jac=True, hess=lambda x: hess(x)[2], method=method
        )
    fcn.vm.set_trans_var(s.x)
    xn = s.x
    ndf = s.x.shape[0]
    min_nll = s.fun
    if not s.success:
        if np.min(np.abs(s.jac)) < gtol:
            s.success = True
            s.message = s.message + "\n But gradients allow"
    success = s.success
    print(s)
    # fcn.vm.set_all(xn)
    params = fcn.get_params()
    return FitResult(
        params, fcn, min_nll, ndf=ndf, success=success, hess_inv=None
    )


def except_result(fcn, ndf):
    params = fcn.vm.get_all_dic()
    return FitResult(
        params, fcn, float(fcn.cached_nll), ndf=ndf, success=False
    )


# import pymultinest
def fit_multinest(fcn):
    pass


class FitResult(object):
    def __init__(
        self, params, fcn, min_nll, ndf=0, success=True, hess_inv=None
    ):
        self.params = params
        self.error = {}
        self.fcn = fcn
        self.min_nll = float(min_nll)
        self.ndf = int(ndf)
        self.success = success
        self.hess_inv = hess_inv

    def save_as(self, file_name, save_hess=False):
        s = {
            "value": self.params,
            "error": self.error,
            "status": {
                "success": self.success,
                "NLL": self.min_nll,
                "Ndf": self.ndf,
            },
        }
        if save_hess and self.hess_inv is not None:
            s["free_params"] = [str(i) for i in self.error]
            s["hess_inv"] = [[float(j) for j in i] for i in self.hess_inv]
        with open(file_name, "w") as f:
            json.dump(s, f, indent=2)

    def set_error(self, error):
        self.error = error.copy()
